# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import strawberry.mutation
import strawberry.type
from framework_strawberry._target_schema_sync import (
    Item,
    Library,
    Storage,
    books,
    libraries,
    magazines,
)
from strawberry import Schema, field
from strawberry.asgi import GraphQL
from strawberry.schema.config import StrawberryConfig
from strawberry.types.types import Optional
from testing_support.asgi_testing import AsgiTest

storage = []


async def resolve_hello():
    return "Hello!"


async def resolve_echo(echo: str):
    return echo


async def resolve_library(index: int):
    return libraries[index]


async def resolve_storage_add(string: str):
    storage.append(string)
    return string


async def resolve_storage():
    return [storage.pop()]


async def resolve_error():
    raise RuntimeError("Runtime Error!")


async def resolve_search(contains: str):
    search_books = [b for b in books if contains in b.name]
    search_magazines = [m for m in magazines if contains in m.name]
    return search_books + search_magazines


@strawberry.type
class Query:
    library: Library = field(resolver=resolve_library)
    hello: str = field(resolver=resolve_hello)
    search: List[Item] = field(resolver=resolve_search)
    echo: str = field(resolver=resolve_echo)
    storage: Storage = field(resolver=resolve_storage)
    error: Optional[str] = field(resolver=resolve_error)
    error_non_null: str = field(resolver=resolve_error)


@strawberry.type
class Mutation:
    storage_add: str = strawberry.mutation(resolver=resolve_storage_add)


target_schema = Schema(query=Query, mutation=Mutation, config=StrawberryConfig(auto_camel_case=False))
target_asgi_application = AsgiTest(GraphQL(target_schema))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import functools
import time

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import database_trace
from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import external_trace
from newrelic.api.function_trace import function_trace
from newrelic.api.graphql_trace import graphql_operation_trace, graphql_resolver_trace
from newrelic.api.memcache_trace import memcache_trace
from newrelic.api.message_trace import message_trace

from newrelic.common.async_wrapper import generator_wrapper

from testing_support.fixtures import capture_transaction_metrics
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

trace_metric_cases = [
    (functools.partial(function_trace, name="simple_gen"), "Function/simple_gen"),
    (functools.partial(external_trace, library="lib", url="http://foo.com"), "External/foo.com/lib/"),
    (functools.partial(database_trace, "select * from foo"), "Datastore/statement/None/foo/select"),
    (functools.partial(datastore_trace, "lib", "foo", "bar"), "Datastore/statement/lib/foo/bar"),
    (functools.partial(message_trace, "lib", "op", "typ", "name"), "MessageBroker/lib/typ/op/Named/name"),
    (functools.partial(memcache_trace, "cmd"), "Memcache/cmd"),
    (functools.partial(graphql_operation_trace), "GraphQL/operation/GraphQL/<unknown>/<anonymous>/<unknown>"),
    (functools.partial(graphql_resolver_trace), "GraphQL/resolve/GraphQL/<unknown>"),
]


@pytest.mark.parametrize("trace,metric", trace_metric_cases)
def test_automatic_generator_trace_wrapper(trace, metric):
    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_automatic_generator_trace_wrapper", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    @background_task(name="test_automatic_generator_trace_wrapper")
    def _test():
        @trace()
        def gen():
            time.sleep(0.1)
            yield
            time.sleep(0.1)

        for _ in gen():
            pass

    _test()

    # Check that generators time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.2


@pytest.mark.parametrize("trace,metric", trace_metric_cases)
def test_manual_generator_trace_wrapper(trace, metric):
    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_automatic_generator_trace_wrapper", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    @background_task(name="test_automatic_generator_trace_wrapper")
    def _test():
        @trace(async_wrapper=generator_wrapper)
        def wrapper_func():
            """Function that returns a generator object, obscuring the automatic introspection of async_wrapper()"""
            def gen():
                time.sleep(0.1)
                yield
                time.sleep(0.1)
            return gen()

        for _ in wrapper_func():
            pass

    _test()

    # Check that generators time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.2



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import (
        transient_function_wrapper,
        function_wrapper)


def validate_serverless_data(
        expected_methods=(), forgone_methods=()):

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        payloads = []

        @transient_function_wrapper('newrelic.common.agent_http',
                'ServerlessModeClient.finalize')
        def _capture(wrapped, instance, args, kwargs):
            payload = wrapped(*args, **kwargs)
            payloads.append(payload)
            return payload

        def _validate():
            assert payloads

            for payload in payloads:
                assert 'metric_data' in payload

                for method in expected_methods:
                    assert method in payload

                    # Verify the method is not a byte string
                    assert isinstance(payload[method], (dict, list))

                for method in forgone_methods:
                    assert method not in payload

        capture_wrapped = _capture(wrapped)
        result = capture_wrapped(*args, **kwargs)
        _validate()
        return result

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import elasticsearch.client
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

from conftest import ES_VERSION, ES_SETTINGS


# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

# Metrics

_base_scoped_metrics = [
    ("Datastore/statement/Elasticsearch/_all/cluster.health", 1),
    ("Datastore/statement/Elasticsearch/_all/search", 2),
    ("Datastore/statement/Elasticsearch/address/index", 2),
    ("Datastore/statement/Elasticsearch/address/search", 1),
    ("Datastore/statement/Elasticsearch/contacts/index", 3),
    ("Datastore/statement/Elasticsearch/contacts/indices.refresh", 1),
    ("Datastore/statement/Elasticsearch/contacts/search", 2),
    ("Datastore/statement/Elasticsearch/other/search", 2),
]

_base_rollup_metrics = [
    ("Datastore/operation/Elasticsearch/cluster.health", 1),
    ("Datastore/operation/Elasticsearch/index", 5),
    ("Datastore/operation/Elasticsearch/indices.refresh", 1),
    ("Datastore/operation/Elasticsearch/search", 7),
    ("Datastore/statement/Elasticsearch/_all/cluster.health", 1),
    ("Datastore/statement/Elasticsearch/_all/search", 2),
    ("Datastore/statement/Elasticsearch/address/index", 2),
    ("Datastore/statement/Elasticsearch/address/search", 1),
    ("Datastore/statement/Elasticsearch/contacts/index", 3),
    ("Datastore/statement/Elasticsearch/contacts/indices.refresh", 1),
    ("Datastore/statement/Elasticsearch/contacts/search", 2),
    ("Datastore/statement/Elasticsearch/other/search", 2),
]

# Version support

def is_importable(module_path):
    try:
        __import__(module_path)
        return True
    except ImportError:
        return False


_all_count = 14

if is_importable("elasticsearch.client.cat") or is_importable("elasticsearch._sync.client.cat"):
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/cat.health", 1))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/cat.health", 1))
    _all_count += 1
else:
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/cat.health", None))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/cat.health", None))

if is_importable("elasticsearch.client.nodes") or is_importable("elasticsearch._sync.client.nodes"):
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/nodes.info", 1))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/nodes.info", 1))
    _all_count += 1
else:
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/nodes.info", None))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/nodes.info", None))

if hasattr(elasticsearch.client, "SnapshotClient") and hasattr(elasticsearch.client.SnapshotClient, "status"):
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/snapshot.status", 1))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/snapshot.status", 1))
    _all_count += 1
else:
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/snapshot.status", None))
    _base_rollup_metrics.append(("Datastore/operation/Elasticsearch/snapshot.status", None))

if hasattr(elasticsearch.client.IndicesClient, "status"):
    _base_scoped_metrics.append(("Datastore/statement/Elasticsearch/_all/indices.status", 1))
    _base_rollup_metrics.extend(
        [
            ("Datastore/operation/Elasticsearch/indices.status", 1),
            ("Datastore/statement/Elasticsearch/_all/indices.status", 1),
        ]
    )
    _all_count += 1
else:
    _base_scoped_metrics.append(("Datastore/operation/Elasticsearch/indices.status", None))
    _base_rollup_metrics.extend(
        [
            ("Datastore/operation/Elasticsearch/indices.status", None),
            ("Datastore/statement/Elasticsearch/_all/indices.status", None),
        ]
    )

_base_rollup_metrics.extend(
    [
        ("Datastore/all", _all_count),
        ("Datastore/allOther", _all_count),
        ("Datastore/Elasticsearch/all", _all_count),
        ("Datastore/Elasticsearch/allOther", _all_count),
    ]
)

# Instance info

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(ES_SETTINGS["host"])
_port = ES_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Elasticsearch/%s/%s" % (_host, _port)

_enable_rollup_metrics.append((_instance_metric_name, _all_count))

_disable_rollup_metrics.append((_instance_metric_name, None))

# Query


def _exercise_es_v7(es):
    es.index(index="contacts", doc_type="person", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    es.index(
        index="contacts", doc_type="person", body={"name": "Jessica Coder", "age": 32, "title": "Programmer"}, id=2
    )
    es.index(index="contacts", doc_type="person", body={"name": "Freddy Tester", "age": 29, "title": "Assistant"}, id=3)
    es.indices.refresh("contacts")
    es.index(
        index="address", doc_type="employee", body={"name": "Sherlock", "address": "221B Baker Street, London"}, id=1
    )
    es.index(
        index="address",
        doc_type="employee",
        body={"name": "Bilbo", "address": "Bag End, Bagshot row, Hobbiton, Shire"},
        id=2,
    )
    es.search(index="contacts", q="name:Joe")
    es.search(index="contacts", q="name:jessica")
    es.search(index="address", q="name:Sherlock")
    es.search(index=["contacts", "address"], q="name:Bilbo")
    es.search(index="contacts,address", q="name:Bilbo")
    es.search(index="*", q="name:Bilbo")
    es.search(q="name:Bilbo")
    es.cluster.health()

    if hasattr(es, "cat"):
        es.cat.health()
    if hasattr(es, "nodes"):
        es.nodes.info()
    if hasattr(es, "snapshot") and hasattr(es.snapshot, "status"):
        es.snapshot.status()
    if hasattr(es.indices, "status"):
        es.indices.status()


def _exercise_es_v8(es):
    es.index(index="contacts", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    es.index(index="contacts", body={"name": "Jessica Coder", "age": 32, "title": "Programmer"}, id=2)
    es.index(index="contacts", body={"name": "Freddy Tester", "age": 29, "title": "Assistant"}, id=3)
    es.indices.refresh(index="contacts")
    es.index(index="address", body={"name": "Sherlock", "address": "221B Baker Street, London"}, id=1)
    es.index(index="address", body={"name": "Bilbo", "address": "Bag End, Bagshot row, Hobbiton, Shire"}, id=2)
    es.search(index="contacts", q="name:Joe")
    es.search(index="contacts", q="name:jessica")
    es.search(index="address", q="name:Sherlock")
    es.search(index=["contacts", "address"], q="name:Bilbo")
    es.search(index="contacts,address", q="name:Bilbo")
    es.search(index="*", q="name:Bilbo")
    es.search(q="name:Bilbo")
    es.cluster.health()

    if hasattr(es, "cat"):
        es.cat.health()
    if hasattr(es, "nodes"):
        es.nodes.info()
    if hasattr(es, "snapshot") and hasattr(es.snapshot, "status"):
        es.snapshot.status()
    if hasattr(es.indices, "status"):
        es.indices.status()


_exercise_es = _exercise_es_v7 if ES_VERSION < (8, 0, 0) else _exercise_es_v8


# Test

@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "test_elasticsearch:test_elasticsearch_operation_disabled",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@override_application_settings(_disable_instance_settings)
@background_task()
def test_elasticsearch_operation_disabled(client):
    _exercise_es(client)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "test_elasticsearch:test_elasticsearch_operation_enabled",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@override_application_settings(_enable_instance_settings)
@background_task()
def test_elasticsearch_operation_enabled(client):
    _exercise_es(client)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

import flask
import flask.views

app = flask.Flask(__name__)

class TestView(flask.views.View):
    def dispatch_request(self):
        return 'VIEW RESPONSE'

class TestMethodView(flask.views.MethodView):
    def get(self):
        return 'METHODVIEW GET RESPONSE'

    def post(self):
        return 'METHODVIEW POST RESPONSE'

app.add_url_rule('/view',
        view_func=TestView.as_view('test_view'))
app.add_url_rule('/methodview',
        view_func=TestMethodView.as_view('test_methodview'))

_test_application = webtest.TestApp(app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import psycopg2
from psycopg2 import extensions as ext
from utils import DB_SETTINGS
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.hooks.database_psycopg2 import wrapper_psycopg2_as_string


class TestCompatability(object):
    def as_string(self, giraffe, lion, tiger=None):
        assert isinstance(giraffe, ext.cursor)
        return "PASS"

wrap_function_wrapper(__name__, 'TestCompatability.as_string',
    wrapper_psycopg2_as_string)


@pytest.fixture(scope='module')
def conn():
    conn = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'])
    yield conn
    conn.close()


def test_forward_compat_args(conn):
    cursor = conn.cursor()
    query = TestCompatability()
    result = query.as_string(cursor, 'giraffe-nomming-leaves')
    assert result == 'PASS'


def test_forward_compat_kwargs(conn):
    cursor = conn.cursor()
    query = TestCompatability()
    result = query.as_string(cursor, lion='eats tiger', tiger='eats giraffe')
    assert result == 'PASS'



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aiohttp
import pytest
from testing_support.fixtures import override_generic_settings
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.core.config import global_settings

version_info = tuple(int(_) for _ in aiohttp.__version__.split(".")[:2])


async def middleware_factory(app, handler):
    async def middleware_handler(request):
        response = await handler(request)
        return response

    return middleware_handler


middleware_tests = [
    (middleware_factory, "Function/test_middleware:" "middleware_factory.<locals>.middleware_handler"),
]


if version_info >= (3, 0):

    @aiohttp.web.middleware
    async def new_style_middleware(request, handler):
        response = await handler(request)
        return response

    middleware_tests.append(
        (new_style_middleware, "Function/test_middleware:new_style_middleware"),
    )


@pytest.mark.parametrize("nr_enabled", [True, False])
@pytest.mark.parametrize("middleware,metric", middleware_tests)
def test_middleware(nr_enabled, aiohttp_app, middleware, metric):
    async def fetch():
        resp = await aiohttp_app.client.request("GET", "/coro")
        assert resp.status == 200
        text = await resp.text()
        assert "Hello Aiohttp!" in text
        return resp

    def _test():
        aiohttp_app.loop.run_until_complete(fetch())

    if nr_enabled:
        scoped_metrics = [
            ("Function/_target_application:index", 1),
            (metric, 1),
        ]

        rollup_metrics = [
            ("Function/_target_application:index", 1),
            (metric, 1),
            ("Python/Framework/aiohttp/%s" % aiohttp.__version__, 1),
        ]

        _test = validate_transaction_metrics(
            "_target_application:index", scoped_metrics=scoped_metrics, rollup_metrics=rollup_metrics
        )(_test)
        _test = validate_code_level_metrics("_target_application", "index")(_test)

        func_name = metric.split("/")[1].replace(":", ".").split(".")
        namespace, func_name = ".".join(func_name[:-1]), func_name[-1]
        _test = validate_code_level_metrics(namespace, func_name)(_test)
    else:
        settings = global_settings()

        _test = override_generic_settings(settings, {"enabled": False})(_test)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import newrelic.api.import_hook as import_hook
import newrelic.packages.six as six
import pytest

from newrelic.config import _module_function_glob

# a dummy hook just to be able to register hooks for modules
def hook(*args, **kwargs):
    pass


def test_import_hook_finder(monkeypatch):
    """
    This asserts the behavior of ImportHookFinder.find_module. It behaves
    differently depending on whether or not the module it is looking for
    exists, has been registered with an import hook, and across different
    python versions.
    """
    finder = import_hook.ImportHookFinder()

    # Override the registered import hooks for the scope of this test
    registered_hooks = {
        "registered_but_does_not_exist": hook,
        "newrelic.api": hook,
    }
    monkeypatch.setattr(import_hook, "_import_hooks", registered_hooks)

    # Finding a module that does not exist and is not registered returns None.
    module = finder.find_module("module_does_not_exist")
    assert module is None

    # Finding a module that does not exist and is registered behaves
    # differently on python 2 vs python 3.
    if six.PY2:
        with pytest.raises(ImportError):
            module = finder.find_module("registered_but_does_not_exist")
    else:
        module = finder.find_module("registered_but_does_not_exist")
        assert module is None

    # Finding a module that exists, but is not registered returns None.
    module = finder.find_module("newrelic")
    assert module is None

    # Finding a module that exists, and is registered, finds that module.
    module = finder.find_module("newrelic.api")
    assert module is not None


@pytest.mark.parametrize("input,expected", [
    ("*", {"run", "A.run", "B.run"}),
    ("NotFound.*", set()),
    ("r*", {"run"}),
    ("*.run", {"A.run", "B.run"}),
    ("A.*", {"A.run"}),
    ("[A,B].run", {"A.run", "B.run"}),
    ("B.r?n", {"B.run"}),
    ("*.RUN", set()),  # Check for case insensitivity issues
])
def test_module_function_globbing(input, expected):
    """This asserts the behavior of filename style globbing on modules."""
    import _test_import_hook as module
    
    result = set(_module_function_glob(module, input))
    assert result == expected, (result, expected)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import warnings

import pytest
from testing_support.validators.validate_function_called import validate_function_called

from newrelic.common.package_version_utils import (
    NULL_VERSIONS,
    VERSION_ATTRS,
    _get_package_version,
    get_package_version,
    get_package_version_tuple,
)
from newrelic.packages import six

# Notes:
# importlib.metadata was a provisional addition to the std library in PY38 and PY39
# while pkg_resources was deprecated.
# importlib.metadata is no longer provisional in PY310+.  It added some attributes
# such as distribution_packages and removed pkg_resources.

IS_PY38_PLUS = sys.version_info[:2] >= (3, 8)
IS_PY310_PLUS = sys.version_info[:2] >= (3, 10)
SKIP_IF_NOT_IMPORTLIB_METADATA = pytest.mark.skipif(not IS_PY38_PLUS, reason="importlib.metadata is not supported.")
SKIP_IF_IMPORTLIB_METADATA = pytest.mark.skipif(
    IS_PY38_PLUS, reason="importlib.metadata is preferred over pkg_resources."
)
SKIP_IF_NOT_PY310_PLUS = pytest.mark.skipif(not IS_PY310_PLUS, reason="These features were added in 3.10+")


@pytest.fixture(scope="function", autouse=True)
def patched_pytest_module(monkeypatch):
    for attr in VERSION_ATTRS:
        if hasattr(pytest, attr):
            monkeypatch.delattr(pytest, attr)

    yield pytest


@pytest.fixture(scope="function", autouse=True)
def cleared_package_version_cache():
    """Ensure cache is empty before every test to exercise code paths."""
    _get_package_version.cache_clear()


# This test only works on Python 3.7
@SKIP_IF_IMPORTLIB_METADATA
@pytest.mark.parametrize(
    "attr,value,expected_value",
    (
        ("version", "1.2.3.4", "1.2.3.4"),
        ("__version__", "1.3.5rc2", "1.3.5rc2"),
        ("__version_tuple__", (3, 5, 8), "3.5.8"),
        ("version_tuple", [3, 1, "0b2"], "3.1.0b2"),
    ),
)
def test_get_package_version(monkeypatch, attr, value, expected_value):
    # There is no file/module here, so we monkeypatch
    # pytest instead for our purposes
    monkeypatch.setattr(pytest, attr, value, raising=False)
    version = get_package_version("pytest")
    assert version == expected_value


# This test only works on Python 3.7
@SKIP_IF_IMPORTLIB_METADATA
def test_skips_version_callables(monkeypatch):
    # There is no file/module here, so we monkeypatch
    # pytest instead for our purposes
    monkeypatch.setattr(pytest, "version", lambda x: "1.2.3.4", raising=False)
    monkeypatch.setattr(pytest, "version_tuple", [3, 1, "0b2"], raising=False)

    version = get_package_version("pytest")

    assert version == "3.1.0b2"


# This test only works on Python 3.7
@SKIP_IF_IMPORTLIB_METADATA
@pytest.mark.parametrize(
    "attr,value,expected_value",
    (
        ("version", "1.2.3.4", (1, 2, 3, 4)),
        ("__version__", "1.3.5rc2", (1, 3, "5rc2")),
        ("__version_tuple__", (3, 5, 8), (3, 5, 8)),
        ("version_tuple", [3, 1, "0b2"], (3, 1, "0b2")),
    ),
)
def test_get_package_version_tuple(monkeypatch, attr, value, expected_value):
    # There is no file/module here, so we monkeypatch
    # pytest instead for our purposes
    monkeypatch.setattr(pytest, attr, value, raising=False)
    version = get_package_version_tuple("pytest")
    assert version == expected_value


@SKIP_IF_NOT_IMPORTLIB_METADATA
@validate_function_called("importlib.metadata", "version")
def test_importlib_metadata():
    version = get_package_version("pytest")
    assert version not in NULL_VERSIONS, version


@SKIP_IF_NOT_PY310_PLUS
@validate_function_called("importlib.metadata", "packages_distributions")
def test_mapping_import_to_distribution_packages():
    version = get_package_version("pytest")
    assert version not in NULL_VERSIONS, version


@SKIP_IF_IMPORTLIB_METADATA
@validate_function_called("pkg_resources", "get_distribution")
def test_pkg_resources_metadata():
    version = get_package_version("pytest")
    assert version not in NULL_VERSIONS, version


def _getattr_deprecation_warning(attr):
    if attr == "__version__":
        warnings.warn("Testing deprecation warnings.", DeprecationWarning)
        return "3.2.1"
    else:
        raise NotImplementedError()


@pytest.mark.skipif(six.PY2, reason="Can't add Deprecation in __version__ in Python 2.")
def test_deprecation_warning_suppression(monkeypatch, recwarn):
    # Add fake module to be deleted later
    monkeypatch.setattr(pytest, "__getattr__", _getattr_deprecation_warning, raising=False)

    assert get_package_version("pytest") == "3.2.1"

    assert not recwarn.list, "Warnings not suppressed."


def test_version_caching(monkeypatch):
    # Add fake module to be deleted later
    sys.modules["mymodule"] = sys.modules["pytest"]
    monkeypatch.setattr(pytest, "__version__", "1.0.0", raising=False)
    version = get_package_version("mymodule")
    assert version not in NULL_VERSIONS, version

    # Ensure after deleting that the call to _get_package_version still completes because of caching
    del sys.modules["mymodule"]
    version = get_package_version("mymodule")
    assert version not in NULL_VERSIONS, version



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
)

from ._target_schema_sync import books, libraries, magazines

storage = []


async def resolve_library(parent, info, index):
    return libraries[index]


async def resolve_storage_add(parent, info, string):
    storage.append(string)
    return string


async def resolve_storage(parent, info):
    return [storage.pop()]


async def resolve_search(parent, info, contains):
    search_books = [b for b in books if contains in b["name"]]
    search_magazines = [m for m in magazines if contains in m["name"]]
    return search_books + search_magazines


Author = GraphQLObjectType(
    "Author",
    {
        "first_name": GraphQLField(GraphQLString),
        "last_name": GraphQLField(GraphQLString),
    },
)

Book = GraphQLObjectType(
    "Book",
    {
        "id": GraphQLField(GraphQLInt),
        "name": GraphQLField(GraphQLString),
        "isbn": GraphQLField(GraphQLString),
        "author": GraphQLField(Author),
        "branch": GraphQLField(GraphQLString),
    },
)

Magazine = GraphQLObjectType(
    "Magazine",
    {
        "id": GraphQLField(GraphQLInt),
        "name": GraphQLField(GraphQLString),
        "issue": GraphQLField(GraphQLInt),
        "branch": GraphQLField(GraphQLString),
    },
)


Library = GraphQLObjectType(
    "Library",
    {
        "id": GraphQLField(GraphQLInt),
        "branch": GraphQLField(GraphQLString),
        "book": GraphQLField(GraphQLList(Book)),
        "magazine": GraphQLField(GraphQLList(Magazine)),
    },
)

Storage = GraphQLList(GraphQLString)


async def resolve_hello(root, info):
    return "Hello!"


async def resolve_echo(root, info, echo):
    return echo


async def resolve_error(root, info):
    raise RuntimeError("Runtime Error!")


hello_field = GraphQLField(GraphQLString, resolve=resolve_hello)
library_field = GraphQLField(
    Library,
    resolve=resolve_library,
    args={"index": GraphQLArgument(GraphQLNonNull(GraphQLInt))},
)
search_field = GraphQLField(
    GraphQLList(GraphQLUnionType("Item", (Book, Magazine), resolve_type=resolve_search)),
    args={"contains": GraphQLArgument(GraphQLNonNull(GraphQLString))},
)
echo_field = GraphQLField(
    GraphQLString,
    resolve=resolve_echo,
    args={"echo": GraphQLArgument(GraphQLNonNull(GraphQLString))},
)
storage_field = GraphQLField(
    Storage,
    resolve=resolve_storage,
)
storage_add_field = GraphQLField(
    GraphQLString,
    resolve=resolve_storage_add,
    args={"string": GraphQLArgument(GraphQLNonNull(GraphQLString))},
)
error_field = GraphQLField(GraphQLString, resolve=resolve_error)
error_non_null_field = GraphQLField(GraphQLNonNull(GraphQLString), resolve=resolve_error)
error_middleware_field = GraphQLField(GraphQLString, resolve=resolve_hello)

query = GraphQLObjectType(
    name="Query",
    fields={
        "hello": hello_field,
        "library": library_field,
        "search": search_field,
        "echo": echo_field,
        "storage": storage_field,
        "error": error_field,
        "error_non_null": error_non_null_field,
        "error_middleware": error_middleware_field,
    },
)

mutation = GraphQLObjectType(
    name="Mutation",
    fields={
        "storage_add": storage_add_field,
    },
)

target_schema = GraphQLSchema(query=query, mutation=mutation)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
import starlette
from testing_support.fixtures import override_ignore_status_codes
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.object_names import callable_name
from newrelic.common.package_version_utils import get_package_version_tuple

starlette_version = get_package_version_tuple("starlette")[:3]


@pytest.fixture(scope="session")
def target_application():
    import _test_application

    return _test_application.target_application


FRAMEWORK_METRIC = ("Python/Framework/Starlette/%s" % starlette.__version__, 1)

if starlette_version >= (0, 20, 1):
    DEFAULT_MIDDLEWARE_METRICS = [
        ("Function/starlette.middleware.errors:ServerErrorMiddleware.__call__", 1),
        ("Function/starlette.middleware.exceptions:ExceptionMiddleware.__call__", 1),
    ]
else:
    DEFAULT_MIDDLEWARE_METRICS = [
        ("Function/starlette.middleware.errors:ServerErrorMiddleware.__call__", 1),
        ("Function/starlette.exceptions:ExceptionMiddleware.__call__", 1),
    ]

MIDDLEWARE_METRICS = [
    ("Function/_test_application:middleware_factory.<locals>.middleware", 2),
    ("Function/_test_application:middleware_decorator", 1),
] + DEFAULT_MIDDLEWARE_METRICS


@pytest.mark.parametrize("app_name", ("no_error_handler",))
@validate_transaction_metrics(
    "_test_application:index",
    scoped_metrics=MIDDLEWARE_METRICS + [("Function/_test_application:index", 1)],
    rollup_metrics=[FRAMEWORK_METRIC],
)
@validate_code_level_metrics("_test_application", "index")
@validate_code_level_metrics("_test_application.middleware_factory.<locals>", "middleware", count=2)
def test_application_index(target_application, app_name):
    app = target_application[app_name]
    response = app.get("/index")
    assert response.status == 200


@pytest.mark.parametrize("app_name", ("no_error_handler",))
@validate_transaction_metrics(
    "_test_application:non_async",
    scoped_metrics=MIDDLEWARE_METRICS + [("Function/_test_application:non_async", 1)],
    rollup_metrics=[FRAMEWORK_METRIC],
)
@validate_code_level_metrics("_test_application", "non_async")
@validate_code_level_metrics("_test_application.middleware_factory.<locals>", "middleware", count=2)
def test_application_non_async(target_application, app_name):
    app = target_application[app_name]
    response = app.get("/non_async")
    assert response.status == 200


# Starting in Starlette v0.20.1, the ExceptionMiddleware class
# has been moved to the starlette.middleware.exceptions from
# starlette.exceptions
version_tweak_string = ".middleware" if starlette_version >= (0, 20, 1) else ""

DEFAULT_MIDDLEWARE_METRICS = [
    ("Function/starlette.middleware.errors:ServerErrorMiddleware.__call__", 1),
    ("Function/starlette%s.exceptions:ExceptionMiddleware.__call__" % version_tweak_string, 1),
]

middleware_test = (
    ("no_error_handler", "starlette%s.exceptions:ExceptionMiddleware.__call__" % version_tweak_string),
    (
        "non_async_error_handler_no_middleware",
        "starlette%s.exceptions:ExceptionMiddleware.__call__" % version_tweak_string,
    ),
)


@pytest.mark.parametrize(
    "app_name, transaction_name",
    middleware_test,
)
def test_application_nonexistent_route(target_application, app_name, transaction_name):
    @validate_transaction_metrics(
        transaction_name,
        scoped_metrics=[("Function/" + transaction_name, 1)],
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    def _test():
        app = target_application[app_name]
        response = app.get("/nonexistent_route")
        assert response.status == 404

    _test()


@pytest.mark.parametrize("app_name", ("no_error_handler",))
def test_exception_in_middleware(target_application, app_name):
    app = target_application[app_name]

    # Starlette >=0.15 and <0.17 raises an exception group instead of reraising the ValueError
    # This only occurs on Python versions >=3.8
    if sys.version_info[0:2] > (3, 7) and starlette_version >= (0, 15, 0) and starlette_version < (0, 17, 0):
        from anyio._backends._asyncio import ExceptionGroup

        exc_type = ExceptionGroup
    else:
        exc_type = ValueError

    @validate_transaction_metrics(
        "_test_application:middleware_factory.<locals>.middleware",
        scoped_metrics=[("Function/_test_application:middleware_factory.<locals>.middleware", 1)],
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    @validate_transaction_errors(errors=[callable_name(exc_type)])
    def _test():
        with pytest.raises(exc_type):  # Later versions of starlette
            app.get("/crash_me_now")

    _test()


@pytest.mark.parametrize(
    "app_name,transaction_name,path,scoped_metrics",
    (
        (
            "non_async_error_handler_no_middleware",
            "_test_application:runtime_error",
            "/runtime_error",
            [],
        ),
        (
            "async_error_handler_no_middleware",
            "_test_application:runtime_error",
            "/runtime_error",
            [("Function/_test_application:async_error_handler", 1)],
        ),
        (
            "no_middleware",
            "_test_application:runtime_error",
            "/runtime_error",
            [
                (
                    "Function/starlette.middleware.errors:ServerErrorMiddleware.error_response",
                    1,
                )
            ],
        ),
        (
            "debug_no_middleware",
            "_test_application:runtime_error",
            "/runtime_error",
            [
                (
                    "Function/starlette.middleware.errors:ServerErrorMiddleware.debug_response",
                    1,
                )
            ],
        ),
        ("no_middleware", "_test_application:CustomRoute", "/raw_runtime_error", []),
    ),
)
@validate_transaction_errors(errors=["builtins:RuntimeError"])
def test_server_error_middleware(target_application, app_name, transaction_name, path, scoped_metrics):
    @validate_transaction_metrics(
        transaction_name,
        scoped_metrics=scoped_metrics + [("Function/_test_application:runtime_error", 1)] + DEFAULT_MIDDLEWARE_METRICS,
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    def _test():
        app = target_application[app_name]
        with pytest.raises(RuntimeError):
            app.get(path)

    _test()


@pytest.mark.parametrize(
    "app_name,transaction_name,path,error",
    (
        (
            "async_error_handler_no_middleware",
            "_test_application:handled_error",
            "/handled_error",
            "_test_application:HandledError",
        ),
        (
            "non_async_error_handler_no_middleware",
            "_test_application:non_async_handled_error",
            "/non_async_handled_error",
            "_test_application:NonAsyncHandledError",
        ),
    ),
)
def test_application_handled_error(target_application, app_name, transaction_name, path, error):
    @validate_transaction_errors(errors=[error])
    @validate_transaction_metrics(
        transaction_name,
        scoped_metrics=[("Function/" + transaction_name, 1)],
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    def _test():
        app = target_application[app_name]
        response = app.get(path)
        assert response.status == 500

    _test()


@pytest.mark.parametrize(
    "app_name,transaction_name,path",
    (
        (
            "async_error_handler_no_middleware",
            "_test_application:handled_error",
            "/handled_error",
        ),
        (
            "non_async_error_handler_no_middleware",
            "_test_application:non_async_handled_error",
            "/non_async_handled_error",
        ),
    ),
)
@override_ignore_status_codes(set((500,)))
def test_application_ignored_error(target_application, app_name, transaction_name, path):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        transaction_name,
        scoped_metrics=[("Function/" + transaction_name, 1)],
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    def _test():
        app = target_application[app_name]
        response = app.get(path)
        assert response.status == 500

    _test()


middleware_test_exception = (
    (
        "no_middleware",
        [("Function/starlette%s.exceptions:ExceptionMiddleware.http_exception" % version_tweak_string, 1)],
    ),
    (
        "teapot_exception_handler_no_middleware",
        [("Function/_test_application:teapot_handler", 1)],
    ),
)


@pytest.mark.parametrize("app_name,scoped_metrics", middleware_test_exception)
def test_starlette_http_exception(target_application, app_name, scoped_metrics):
    @validate_transaction_errors(errors=["starlette.exceptions:HTTPException"])
    @validate_transaction_metrics(
        "_test_application:teapot",
        scoped_metrics=scoped_metrics + DEFAULT_MIDDLEWARE_METRICS,
        rollup_metrics=[FRAMEWORK_METRIC],
    )
    def _test():
        app = target_application[app_name]
        response = app.get("/418")
        assert response.status == 418

    _test()


@pytest.mark.parametrize("app_name", ("no_middleware",))
@validate_transaction_errors(errors=["builtins:RuntimeError"])
@validate_transaction_metrics("_test_application:CustomRoute", rollup_metrics=[FRAMEWORK_METRIC])
def test_starlette_http_exception_after_response_start(target_application, app_name):
    app = target_application[app_name]
    with pytest.raises(RuntimeError):
        app.get("/raw_http_error")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from newrelic.common.object_wrapper import (
    function_wrapper,
    transient_function_wrapper,
)
from testing_support.fixtures import catch_background_exceptions

def validate_transaction_metrics(
    name,
    group="Function",
    background_task=False,
    scoped_metrics=None,
    rollup_metrics=None,
    custom_metrics=None,
    dimensional_metrics=None,
    index=-1,
):
    scoped_metrics = scoped_metrics or []
    rollup_metrics = rollup_metrics or []
    custom_metrics = custom_metrics or []
    dimensional_metrics = dimensional_metrics or []

    if background_task:
        unscoped_metrics = [
            "OtherTransaction/all",
            "OtherTransaction/%s/%s" % (group, name),
            "OtherTransactionTotalTime",
            "OtherTransactionTotalTime/%s/%s" % (group, name),
        ]
        transaction_scope_name = "OtherTransaction/%s/%s" % (group, name)
    else:
        unscoped_metrics = [
            "WebTransaction",
            "WebTransaction/%s/%s" % (group, name),
            "WebTransactionTotalTime",
            "WebTransactionTotalTime/%s/%s" % (group, name),
            "HttpDispatcher",
        ]
        transaction_scope_name = "WebTransaction/%s/%s" % (group, name)

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        record_transaction_called = []
        recorded_metrics = []
        recorded_dimensional_metrics = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        @catch_background_exceptions
        def _validate_transaction_metrics(wrapped, instance, args, kwargs):
            record_transaction_called.append(True)
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            else:
                metrics = instance.stats_table
                # Record a copy of the metric value so that the values aren't
                # merged in the future
                _metrics = {}
                for k, v in metrics.items():
                    _metrics[k] = copy.copy(v)
                recorded_metrics.append(_metrics)

                metrics = instance.dimensional_stats_table.metrics()
                # Record a copy of the metric value so that the values aren't
                # merged in the future
                _metrics = {}
                for k, v in metrics:
                    _metrics[k] = copy.copy(v)
                recorded_dimensional_metrics.append(_metrics)

            return result

        def _validate(metrics, name, scope, count):
            key = (name, scope)

            if isinstance(scope, str):
                # Normal metric lookup
                metric = metrics.get(key)
            else:
                # Dimensional metric lookup
                metric_container = metrics.get(name, {})
                metric = metric_container.get(scope)

            def _metrics_table():
                out = [""]
                out.append("Expected: {0}: {1}".format(key, count))
                for metric_key, metric_container in metrics.items():
                    if isinstance(metric_container, dict):
                        for metric_tags, metric_value in metric_container.items():
                            out.append("{0}: {1}".format((metric_key, metric_tags), metric_value[0]))
                    else:
                        out.append("{0}: {1}".format(metric_key, metric_container[0]))
                return "\n".join(out)

            def _metric_details():
                return "metric=%r, count=%r" % (key, metric.call_count)

            if count is not None:
                assert metric is not None, _metrics_table()
                if count == "present":
                    assert metric.call_count > 0, _metric_details()
                else:
                    assert metric.call_count == count, _metric_details()

                assert metric.total_call_time >= 0, (key, metric)
                assert metric.total_exclusive_call_time >= 0, (key, metric)
                assert metric.min_call_time >= 0, (key, metric)
                assert metric.sum_of_squares >= 0, (key, metric)

            else:
                assert metric is None, _metrics_table()

        _new_wrapper = _validate_transaction_metrics(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert record_transaction_called
        metrics = recorded_metrics[index]
        captured_dimensional_metrics = recorded_dimensional_metrics[index]

        record_transaction_called[:] = []
        recorded_metrics[:] = []
        recorded_dimensional_metrics[:] = []

        for unscoped_metric in unscoped_metrics:
            _validate(metrics, unscoped_metric, "", 1)

        for scoped_name, scoped_count in scoped_metrics:
            _validate(metrics, scoped_name, transaction_scope_name, scoped_count)

        for rollup_name, rollup_count in rollup_metrics:
            _validate(metrics, rollup_name, "", rollup_count)

        for custom_name, custom_count in custom_metrics:
            _validate(metrics, custom_name, "", custom_count)

        for dimensional_name, dimensional_tags, dimensional_count in dimensional_metrics:
            if isinstance(dimensional_tags, dict):
                dimensional_tags = frozenset(dimensional_tags.items())
            _validate(captured_dimensional_metrics, dimensional_name, dimensional_tags, dimensional_count)

        custom_metric_names = {name for name, _ in custom_metrics}
        for name, _ in metrics:
            if name not in custom_metric_names:
                assert not name.startswith("Supportability/api/"), name

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

pytestmark = pytest.mark.custom_app


def test_custom_handler(app):
    FRAMEWORK_METRIC = 'Python/Framework/Tornado/%s' % app.tornado_version

    @validate_transaction_metrics(
        name='_target_application:CustomApplication',
        rollup_metrics=((FRAMEWORK_METRIC, 1),),
    )
    @validate_code_level_metrics("_target_application", "CustomApplication")
    def _test():
        response = app.fetch('/')
        assert response.code == 200
        assert response.body == b'*'

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs
from testing_support.util import instance_hostname
from utils import DB_SETTINGS

from newrelic.api.background_task import background_task


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

_base_scoped_metrics = (
        ('Function/psycopg2:connect', 1),
        ('Datastore/operation/Postgres/rollback', 1)
)

_base_rollup_metrics = (
        ('Datastore/all', 2),
        ('Datastore/allOther', 2),
        ('Datastore/Postgres/all', 2),
        ('Datastore/Postgres/allOther', 2),
        ('Datastore/operation/Postgres/rollback', 1)
)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Postgres/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 1)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)

# Query

def _exercise_db():
    try:
        with psycopg2.connect(
                database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
                password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
                port=DB_SETTINGS['port']):

            raise RuntimeError('error')
    except RuntimeError:
        pass

# Tests

@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_rollback:test_rollback_on_exception_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception_enable_instance():
    _exercise_db()


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_rollback:test_rollback_on_exception_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception_disable_instance():
    _exercise_db()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from compat import basic_consume
import pika
import six
import os

from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction
from newrelic.api.function_trace import FunctionTrace
from newrelic.common.encoding_utils import DistributedTracePayload

from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

DB_SETTINGS = rabbitmq_settings()[0]

_override_settings = {
    'primary_application_id': '12345',
    'account_id': '33',
    'trusted_account_key': '1',
    'cross_application_tracer.enabled': True,
    'distributed_tracing.enabled': True,
}

_test_distributed_tracing_basic_publish_metrics = [
    ('Supportability/TraceContext/Create/Success', 1),
    ('Supportability/DistributedTrace/CreatePayload/Success', 1),
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/Default', 1),
    ('DurationByCaller/Unknown/Unknown/Unknown/Unknown/all', 1),
    ('DurationByCaller/Unknown/Unknown/Unknown/Unknown/allOther', 1),
]


@validate_transaction_metrics(
        'test_distributed_tracing:do_basic_publish',
        rollup_metrics=_test_distributed_tracing_basic_publish_metrics,
        background_task=True)
@background_task()
def do_basic_publish(channel, QUEUE, properties=None):
    channel.basic_publish(
        exchange='',
        routing_key=QUEUE,
        body='Testing distributed_tracing 123',
        properties=properties,
    )


_test_distributed_tracing_basic_consume_rollup_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/Default', None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/Default', None),
    ('Supportability/DistributedTrace/AcceptPayload/Success', None),
    ('Supportability/TraceContext/Accept/Success', 1),
    ('DurationByCaller/App/33/12345/AMQP/all', 1),
    ('TransportDuration/App/33/12345/AMQP/all', 1),
    ('DurationByCaller/App/33/12345/AMQP/allOther', 1),
    ('TransportDuration/App/33/12345/AMQP/allOther', 1)

]

if six.PY3:
    _consume_txn_name = ('test_distributed_tracing:'
            'test_basic_consume_distributed_tracing_headers.'
            '<locals>.on_receive')
else:
    _consume_txn_name = (
        'test_distributed_tracing:on_receive')


@validate_transaction_metrics(
        _consume_txn_name,
        rollup_metrics=_test_distributed_tracing_basic_consume_rollup_metrics,
        background_task=True,
        group='Message/RabbitMQ/Exchange/Default')
def do_basic_consume(channel):
    channel.start_consuming()


@override_application_settings(_override_settings)
def test_basic_consume_distributed_tracing_headers():
    def on_receive(ch, method, properties, msg):
        headers = properties.headers
        assert headers
        assert 'NewRelicID' not in headers
        assert 'NewRelicTransaction' not in headers
        assert msg == b'Testing distributed_tracing 123'
        txn = current_transaction()

        assert txn
        assert txn._distributed_trace_state
        assert txn.parent_type == 'App'
        assert txn._trace_id.startswith(txn.parent_tx)
        assert txn.parent_span is not None
        assert txn.parent_account == txn.settings.account_id
        assert txn.parent_transport_type == 'AMQP'
        assert txn._priority is not None
        assert txn._sampled is not None

        ch.stop_consuming()

    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        queue_name = 'TESTDT-%s' % os.getpid()
        channel.queue_declare(queue_name, durable=False)

        properties = pika.BasicProperties()
        properties.headers = {'Hello': 'World'}

        try:
            basic_consume(channel, queue_name, on_receive, auto_ack=False)
            do_basic_publish(channel, queue_name, properties=properties)
            do_basic_consume(channel)

        finally:
            channel.queue_delete(queue_name)


_test_distributed_tracing_basic_get_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/Default', None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/Default', 1),
    ('DurationByCaller/Unknown/Unknown/Unknown/Unknown/all', 1),
    ('DurationByCaller/Unknown/Unknown/Unknown/Unknown/allOther', 1)
]


@validate_transaction_metrics(
        'test_distributed_tracing:do_basic_get',
        rollup_metrics=_test_distributed_tracing_basic_get_metrics,
        background_task=True)
@background_task()
def do_basic_get(channel, QUEUE):
    _, properties, msg = channel.basic_get(QUEUE)
    headers = properties.headers

    assert headers
    assert 'NewRelicID' not in headers
    assert 'NewRelicTransaction' not in headers
    assert msg == b'Testing distributed_tracing 123'

    txn = current_transaction()

    assert txn.client_cross_process_id is None
    assert txn.client_account_id is None
    assert txn.client_application_id is None


@override_application_settings(_override_settings)
def test_basic_get_no_distributed_tracing_headers():
    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        queue_name = 'TESTDT-%s' % os.getpid()
        channel.queue_declare(queue_name, durable=False)

        properties = pika.BasicProperties()
        properties.headers = {'Hello': 'World'}

        try:
            do_basic_publish(channel, queue_name, properties=properties)
            do_basic_get(channel, queue_name)
        finally:
            channel.queue_delete(queue_name)


@override_application_settings(_override_settings)
def test_distributed_tracing_sends_produce_id():
    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        queue_name = 'TESTDT-%s' % os.getpid()
        channel.queue_declare(queue_name, durable=False)

        properties = pika.BasicProperties()
        properties.headers = {'Hello': 'World'}

        try:
            @background_task()
            def _publish():
                with FunctionTrace('foo') as trace:
                    channel.basic_publish(
                        exchange='',
                        routing_key=queue_name,
                        body='Testing distributed_tracing 123',
                        properties=properties,
                    )

                return trace

            trace = _publish()

            raw_message = channel.basic_get(queue_name)
        finally:
            channel.queue_delete(queue_name)

        properties = raw_message[1]
        payload = DistributedTracePayload.from_http_safe(
                properties.headers["newrelic"])

        data = payload['d']

        # The payload should NOT contain the function trace ID
        assert data['id'] != trace.guid



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import AIOREDIS_VERSION, SKIPIF_AIOREDIS_V1, SKIPIF_AIOREDIS_V2
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)

from newrelic.api.background_task import background_task


@background_task()
@pytest.mark.parametrize("in_transaction", (True, False))
def test_pipelines_no_harm(client, in_transaction, loop, key):
    async def exercise():
        if AIOREDIS_VERSION >= (2,):
            pipe = client.pipeline(transaction=in_transaction)
        else:
            pipe = client.pipeline()  # Transaction kwarg unsupported

        pipe.set(key, 1)
        return await pipe.execute()

    status = loop.run_until_complete(exercise())
    assert status == [True]


def exercise_transaction_sync(key):
    def _run(pipe):
        pipe.set(key, 1)
    return _run


def exercise_transaction_async(key):
    async def _run(pipe):
        await pipe.set(key, 1)
    return _run


@SKIPIF_AIOREDIS_V1
@pytest.mark.parametrize("exercise", (exercise_transaction_sync, exercise_transaction_async))
@background_task()
def test_transactions_no_harm(client, loop, key, exercise):
    status = loop.run_until_complete(client.transaction(exercise(key)))
    assert status == [True]


@SKIPIF_AIOREDIS_V2
@background_task()
def test_multi_exec_no_harm(client, loop, key):
    async def exercise():
        pipe = client.multi_exec()
        pipe.set(key, "value")
        status = await pipe.execute()
        assert status == [True]

    loop.run_until_complete(exercise())


@SKIPIF_AIOREDIS_V1
@background_task()
def test_pipeline_immediate_execution_no_harm(client, loop, key):
    async def exercise():
        await client.set(key, 1)

        if AIOREDIS_VERSION >= (2,):
            pipe = client.pipeline(transaction=True)
        else:
            pipe = client.pipeline()  # Transaction kwarg unsupported

        async with pipe:
            await pipe.watch(key)
            value = int(await pipe.get(key))
            assert value == 1
            value += 1
            pipe.multi()
            pipe.set(key, value)
            await pipe.execute()

        assert int(await client.get(key)) == 2

    loop.run_until_complete(exercise())


@SKIPIF_AIOREDIS_V1
@background_task()
def test_transaction_immediate_execution_no_harm(client, loop, key):
    async def exercise():
        async def exercise_transaction(pipe):
            value = int(await pipe.get(key))
            assert value == 1
            value += 1
            pipe.multi()
            pipe.set(key, value)
            await pipe.execute()

        await client.set(key, 1)
        status = await client.transaction(exercise_transaction, key)
        assert int(await client.get(key)) == 2

        return status

    status = loop.run_until_complete(exercise())
    assert status == []


@SKIPIF_AIOREDIS_V1
@validate_transaction_errors([])
@background_task()
def test_transaction_watch_error_no_harm(client, loop, key):
    async def exercise():
        async def exercise_transaction(pipe):
            value = int(await pipe.get(key))
            if value == 1:
                # Only run set the first pass, as this runs repeatedly until no watch error is raised.
                await pipe.set(key, 2)

        await client.set(key, 1)
        status = await client.transaction(exercise_transaction, key)

        return status

    status = loop.run_until_complete(exercise())
    assert status == []



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.core import attribute_filter as af
from newrelic.core.config import global_settings, Settings

from testing_support.fixtures import override_application_settings

def _default_settings():
    return {
        'attributes.enabled': True,
        'transaction_events.attributes.enabled': True,
        'transaction_tracer.attributes.enabled': True,
        'error_collector.attributes.enabled': True,
        'browser_monitoring.attributes.enabled': False,
        'attributes.include': [],
        'attributes.exclude': [],
        'transaction_events.attributes.include': [],
        'transaction_events.attributes.exclude': [],
        'transaction_tracer.attributes.include': [],
        'transaction_tracer.attributes.exclude': [],
        'error_collector.attributes.include': [],
        'error_collector.attributes.exclude': [],
        'browser_monitoring.attributes.include': [],
        'browser_monitoring.attributes.exclude': [],
    }

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.join(CURRENT_DIR, 'fixtures', 'attribute_configuration.json')

def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)

_fields = ['testname', 'config', 'input_key', 'input_default_destinations',
           'expected_destinations']

def _parametrize_test(test):
    return tuple([test.get(f, None) for f in _fields])

_attributes_tests = [_parametrize_test(t) for t in _load_tests()]

def destinations_as_int(destinations):
    result = af.DST_NONE
    for d in destinations:
        if d == 'transaction_events':
            result |= af.DST_TRANSACTION_EVENTS
        elif d == 'transaction_tracer':
            result |= af.DST_TRANSACTION_TRACER
        elif d == 'error_collector':
            result |= af.DST_ERROR_COLLECTOR
        elif d == 'browser_monitoring':
            result |= af.DST_BROWSER_MONITORING
    return result

@pytest.mark.parametrize(','.join(_fields), _attributes_tests)
def test_attributes(testname, config, input_key, input_default_destinations,
            expected_destinations):

    settings = _default_settings()
    for k, v in config.items():
        settings[k] = v

    attribute_filter = af.AttributeFilter(settings)
    input_destinations = destinations_as_int(input_default_destinations)

    result = attribute_filter.apply(input_key, input_destinations)
    expected = destinations_as_int(expected_destinations)
    assert result == expected, attribute_filter

_sorting_tests = [
    ('lexicographic', ('a', 1, True), ('ab', 1, True)),
    ('is_include', ('a', 1, True), ('a', 1, False)),
    ('wildcard', ('a*', 1, True), ('a', 1, False)),
    ('wildcard same length', ('a*', 1, True), ('ab', 1, False)),
]

@pytest.mark.parametrize('testname, rule1, rule2', _sorting_tests)
def test_attribute_filter_rule_sort(testname, rule1, rule2):
    first = af.AttributeFilterRule(*rule1)
    second = af.AttributeFilterRule(*rule2)

    assert first < second
    assert first <= second
    assert second > first
    assert second >= first
    assert first != second

_equality_tests = [
    ('is_include', ('a', 1, True), ('a', 1, True)),
    ('not is_include', ('a', 1, False), ('a', 1, False)),
    ('is_wildcard', ('a*', 1, True), ('a*', 1, True)),
]
@pytest.mark.parametrize('testname, rule1, rule2', _equality_tests)
def test_attribute_filter_rule_equality(testname, rule1, rule2):
    first = af.AttributeFilterRule(*rule1)
    second = af.AttributeFilterRule(*rule2)

    assert first == second



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import check_event_attributes

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper


def validate_error_event_attributes(required_params=None, forgone_params=None, exact_attrs=None):
    """Check the error event for attributes, expect only one error to be
    present in the transaction.
    """
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}
    error_data_samples = []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        def _validate_error_event_attributes(wrapped, instance, args, kwargs):
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise

            event_data = instance.error_events
            for sample in event_data:
                error_data_samples.append(sample)

            check_event_attributes(event_data, required_params, forgone_params, exact_attrs)

            return result

        _new_wrapper = _validate_error_event_attributes(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert error_data_samples
        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import time

from testing_support.fixtures import core_application_stats_engine

from newrelic.common.object_wrapper import function_wrapper


def validate_non_transaction_error_event(required_intrinsics=None, num_errors=1, required_user=None, forgone_user=None):
    """Validate error event data for a single error occurring outside of a
    transaction.
    """
    required_intrinsics = required_intrinsics or {}
    required_user = required_user or {}
    forgone_user = forgone_user or []

    @function_wrapper
    def _validate_non_transaction_error_event(wrapped, instace, args, kwargs):

        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            stats = core_application_stats_engine(None)

            assert stats.error_events.num_seen == num_errors
            for event in stats.error_events:

                assert len(event) == 3  # [intrinsic, user, agent attributes]

                intrinsics = event[0]

                # The following attributes are all required, and also the only
                # intrinsic attributes that can be included in an error event
                # recorded outside of a transaction

                assert intrinsics["type"] == "TransactionError"
                assert intrinsics["transactionName"] is None
                assert intrinsics["error.class"] == required_intrinsics["error.class"]
                assert intrinsics["error.message"].startswith(required_intrinsics["error.message"])
                assert intrinsics["error.expected"] == required_intrinsics["error.expected"]
                now = time()
                assert isinstance(intrinsics["timestamp"], int)
                assert intrinsics["timestamp"] <= 1000.0 * now

                user_params = event[1]
                for name, value in required_user.items():
                    assert name in user_params, "name=%r, params=%r" % (name, user_params)
                    assert user_params[name] == value, "name=%r, value=%r, params=%r" % (name, value, user_params)

                for param in forgone_user:
                    assert param not in user_params

        return result

    return _validate_non_transaction_error_event



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jinja2 import Template
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task


@validate_transaction_metrics(
    "test_render",
    background_task=True,
    scoped_metrics=(
        ("Template/Render/<template>", 1),
        ("Template/Compile/<template>", 1),
    ),
)
@background_task(name="test_render")
def test_render():
    template_to_render = Template("hello, {{ name }}!")
    result = template_to_render.render(name="NR")
    assert result == "hello, NR!"


def test_render_outside_txn():
    template_to_render = Template("hello, {{ name }}!")
    result = template_to_render.render(name="NR")
    assert result == "hello, NR!"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import django

BASE_DIR = os.path.dirname(__file__)
DEBUG = True

django_version = django.VERSION

# Make this unique, and don't share it with anybody.
SECRET_KEY = "cookies"

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

middleware = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'middleware.ExceptionTo410Middleware',
)
if django_version[:2] >= (1, 10):
    MIDDLEWARE = middleware
else:
    MIDDLEWARE_CLASSES = middleware

ROOT_URLCONF = 'urls'

TEMPLATE_DIRS = (
    [os.path.join(BASE_DIR, 'templates')]
)

# For Django 1.10 compatibility because TEMPLATE_DIRS is deprecated
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': TEMPLATE_DIRS,
    }
]

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'dummy_app',
    'newrelic.extras.framework_django',
)


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.common.object_wrapper import function_wrapper


@pytest.fixture(scope="function")
def wrapper():
    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        return wrapped(*args, **kwargs)

    return _wrapper


@pytest.fixture(scope="function")
def wrapped_function(wrapper):
    @wrapper
    def wrapped():
        return True

    return wrapped


def test_nr_prefix_attributes(wrapped_function):
    wrapped_function._nr_attr = 1
    vars_ = vars(wrapped_function)

    assert wrapped_function._nr_attr == 1, "_nr_ attributes should be stored on wrapper object and retrievable."
    assert "_nr_attr" not in vars_, "_nr_ attributes should NOT appear in __dict__."


def test_self_prefix_attributes(wrapped_function):
    wrapped_function._self_attr = 1
    vars_ = vars(wrapped_function)

    assert wrapped_function._self_attr == 1, "_self_ attributes should be stored on wrapper object and retrievable."
    assert "_nr_attr" not in vars_, "_self_ attributes should NOT appear in __dict__."


def test_prefixed_attributes_share_namespace(wrapped_function):
    wrapped_function._nr_attr = 1
    wrapped_function._self_attr = 2

    assert (
        wrapped_function._nr_attr == 2
    ), "_nr_ attributes share a namespace with _self_ attributes and should be overwritten."


def test_wrapped_function_attributes(wrapped_function):
    wrapped_function._other_attr = 1
    vars_ = vars(wrapped_function)

    assert wrapped_function._other_attr == 1, "All other attributes should be stored on wrapped object and retrievable."
    assert "_other_attr" in vars_, "Other types of attributes SHOULD appear in __dict__."

    assert wrapped_function()


def test_multiple_wrapper_last_object(wrapper):
    def wrapped():
        pass

    wrapper_1 = wrapper(wrapped)
    wrapper_2 = wrapper(wrapper_1)

    assert wrapper_2._nr_last_object is wrapped, "Last object in chain should be the wrapped function."
    assert wrapper_2._nr_next_object is wrapper_1, "Next object in chain should be the middle function."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

# ===== WriteBatch =====


@pytest.fixture()
def exercise_write_batch(client, collection):
    def _exercise_write_batch():
        docs = [collection.document(str(x)) for x in range(1, 4)]
        batch = client.batch()
        for doc in docs:
            batch.set(doc, {})

        batch.commit()

    return _exercise_write_batch


def test_firestore_write_batch(exercise_write_batch, instance_info):
    _test_scoped_metrics = [
        ("Datastore/operation/Firestore/commit", 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/all", 1),
        ("Datastore/allOther", 1),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 1),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_write_batch",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_write_batch")
    def _test():
        exercise_write_batch()

    _test()


def test_firestore_write_batch_trace_node_datastore_params(exercise_write_batch, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        exercise_write_batch()

    _test()


# ===== BulkWriteBatch =====


@pytest.fixture()
def exercise_bulk_write_batch(client, collection):
    def _exercise_bulk_write_batch():
        from google.cloud.firestore_v1.bulk_batch import BulkWriteBatch

        docs = [collection.document(str(x)) for x in range(1, 4)]
        batch = BulkWriteBatch(client)
        for doc in docs:
            batch.set(doc, {})

        batch.commit()

    return _exercise_bulk_write_batch


def test_firestore_bulk_write_batch(exercise_bulk_write_batch, instance_info):
    _test_scoped_metrics = [
        ("Datastore/operation/Firestore/commit", 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/all", 1),
        ("Datastore/allOther", 1),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 1),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_bulk_write_batch",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_bulk_write_batch")
    def _test():
        exercise_bulk_write_batch()

    _test()


def test_firestore_bulk_write_batch_trace_node_datastore_params(exercise_bulk_write_batch, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        exercise_bulk_write_batch()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys

from newrelic.api.application import application_instance
from newrelic.api.error_trace import wrap_error_trace
from newrelic.api.function_trace import FunctionTraceWrapper
from newrelic.api.message_trace import MessageTrace
from newrelic.api.message_transaction import MessageTransaction
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import function_wrapper, wrap_function_wrapper
from newrelic.common.package_version_utils import get_package_version

_logger = logging.getLogger(__name__)

HEARTBEAT_POLL = "MessageBroker/Kafka/Heartbeat/Poll"
HEARTBEAT_SENT = "MessageBroker/Kafka/Heartbeat/Sent"
HEARTBEAT_FAIL = "MessageBroker/Kafka/Heartbeat/Fail"
HEARTBEAT_RECEIVE = "MessageBroker/Kafka/Heartbeat/Receive"
HEARTBEAT_SESSION_TIMEOUT = "MessageBroker/Kafka/Heartbeat/SessionTimeout"
HEARTBEAT_POLL_TIMEOUT = "MessageBroker/Kafka/Heartbeat/PollTimeout"


def wrap_Producer_produce(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    if transaction is None:
        return wrapped(*args, **kwargs)

    # Binding with a standard function signature does not work properly due to a bug in handling arguments
    # in the underlying C code, where callback=None being specified causes on_delivery=callback to never run.

    # Bind out headers from end of args list
    if len(args) == 8:
        # Take headers off the end of the positional args
        headers = args[7]
        args = args[0:7]
    else:
        headers = kwargs.pop("headers", [])

    # Bind topic off of the beginning of the args list
    if len(args) >= 1:
        topic = args[0]
        args = args[1:]
    else:
        topic = kwargs.pop("topic", None)

    transaction.add_messagebroker_info("Confluent-Kafka", get_package_version("confluent-kafka"))

    with MessageTrace(
        library="Kafka",
        operation="Produce",
        destination_type="Topic",
        destination_name=topic or "Default",
        source=wrapped,
    ):
        dt_headers = {k: v.encode("utf-8") for k, v in MessageTrace.generate_request_headers(transaction)}
        # headers can be a list of tuples or a dict so convert to dict for consistency.
        if headers:
            dt_headers.update(dict(headers))

        try:
            return wrapped(topic, headers=dt_headers, *args, **kwargs)
        except Exception as error:
            # Unwrap kafka errors
            while hasattr(error, "exception"):
                error = error.exception  # pylint: disable=E1101

            _, _, tb = sys.exc_info()
            notice_error((type(error), error, tb))
            tb = None  # Clear reference to prevent reference cycles
            raise


def wrap_Consumer_poll(wrapped, instance, args, kwargs):
    # This wrapper can be called either outside of a transaction, or
    # within the context of an existing transaction.  There are 4
    # possibilities we need to handle: (Note that this is similar to
    # our Pika, Celery, and Kafka-Python instrumentation)
    #
    #   1. Inside an inner wrapper in the DeserializingConsumer
    #
    #       Do nothing. The DeserializingConsumer is double wrapped because
    #       the underlying C implementation is wrapped as well. We need to
    #       detect when the second wrapper is called and ignore it completely
    #       or transactions will be stopped early.
    #
    #   2. In an inactive transaction
    #
    #      If the end_of_transaction() or ignore_transaction() API
    #      calls have been invoked, this iterator may be called in the
    #      context of an inactive transaction. In this case, don't wrap
    #      the iterator in any way. Just run the original iterator.
    #
    #   3. In an active transaction
    #
    #      Do nothing.
    #
    #   4. Outside of a transaction
    #
    #      Since it's not running inside of an existing transaction, we
    #      want to create a new background transaction for it.

    # Step 1: Stop existing transactions
    if hasattr(instance, "_nr_transaction") and not instance._nr_transaction.stopped:
        instance._nr_transaction.__exit__(*sys.exc_info())

    # Step 2: Poll for records
    try:
        record = wrapped(*args, **kwargs)
    except Exception as e:
        if current_transaction():
            notice_error()
        else:
            notice_error(application=application_instance(activate=False))
        raise

    # Step 3: Start new transaction for received record
    if record:
        library = "Kafka"
        destination_type = "Topic"
        destination_name = record.topic()
        received_bytes = len(str(record.value()).encode("utf-8"))
        message_count = 1

        headers = record.headers()
        headers = dict(headers) if headers else {}

        transaction = current_transaction(active_only=False)
        if not transaction:
            transaction = MessageTransaction(
                application=application_instance(),
                library=library,
                destination_type=destination_type,
                destination_name=destination_name,
                headers=headers,
                transport_type="Kafka",
                routing_key=record.key(),
                source=wrapped,
            )
            instance._nr_transaction = transaction
            transaction.__enter__()  # pylint: disable=C2801

            transaction._add_agent_attribute("kafka.consume.byteCount", received_bytes)

        transaction = current_transaction()

        if transaction:  # If there is an active transaction now.
            # Add metrics whether or not a transaction was already active, or one was just started.
            # Don't add metrics if there was an inactive transaction.
            # Name the metrics using the same format as the transaction, but in case the active transaction
            # was an existing one and not a message transaction, reproduce the naming logic here.
            group = "Message/%s/%s" % (library, destination_type)
            name = "Named/%s" % destination_name
            transaction.record_custom_metric("%s/%s/Received/Bytes" % (group, name), received_bytes)
            transaction.record_custom_metric("%s/%s/Received/Messages" % (group, name), message_count)
            transaction.add_messagebroker_info("Confluent-Kafka", get_package_version("confluent-kafka"))

    return record


def wrap_DeserializingConsumer_poll(wrapped, instance, args, kwargs):
    try:
        return wrapped(*args, **kwargs)
    except Exception:
        notice_error()

        # Stop existing transactions
        if hasattr(instance, "_nr_transaction") and not instance._nr_transaction.stopped:
            instance._nr_transaction.__exit__(*sys.exc_info())

        raise


def wrap_serializer(serializer_name, group_prefix):
    @function_wrapper
    def _wrap_serializer(wrapped, instance, args, kwargs):
        if not current_transaction():
            return wrapped(*args, **kwargs)

        topic = args[1].topic
        group = "%s/Kafka/Topic" % group_prefix
        name = "Named/%s/%s" % (topic, serializer_name)

        return FunctionTraceWrapper(wrapped, name=name, group=group)(*args, **kwargs)

    return _wrap_serializer


def wrap_SerializingProducer_init(wrapped, instance, args, kwargs):
    wrapped(*args, **kwargs)

    if hasattr(instance, "_key_serializer") and callable(instance._key_serializer):
        instance._key_serializer = wrap_serializer("Serialization/Key", "MessageBroker")(instance._key_serializer)

    if hasattr(instance, "_value_serializer") and callable(instance._value_serializer):
        instance._value_serializer = wrap_serializer("Serialization/Value", "MessageBroker")(instance._value_serializer)


def wrap_DeserializingConsumer_init(wrapped, instance, args, kwargs):
    wrapped(*args, **kwargs)

    if hasattr(instance, "_key_deserializer") and callable(instance._key_deserializer):
        instance._key_deserializer = wrap_serializer("Deserialization/Key", "Message")(instance._key_deserializer)

    if hasattr(instance, "_value_deserializer") and callable(instance._value_deserializer):
        instance._value_deserializer = wrap_serializer("Deserialization/Value", "Message")(instance._value_deserializer)


def wrap_immutable_class(module, class_name):
    # Wrap immutable binary extension class with a mutable Python subclass
    new_class = type(class_name, (getattr(module, class_name),), {})
    setattr(module, class_name, new_class)
    return new_class


def instrument_confluentkafka_cimpl(module):
    if hasattr(module, "Producer"):
        wrap_immutable_class(module, "Producer")
        wrap_function_wrapper(module, "Producer.produce", wrap_Producer_produce)

    if hasattr(module, "Consumer"):
        wrap_immutable_class(module, "Consumer")
        wrap_function_wrapper(module, "Consumer.poll", wrap_Consumer_poll)


def instrument_confluentkafka_serializing_producer(module):
    if hasattr(module, "SerializingProducer"):
        wrap_function_wrapper(module, "SerializingProducer.__init__", wrap_SerializingProducer_init)
        wrap_error_trace(module, "SerializingProducer.produce")


def instrument_confluentkafka_deserializing_consumer(module):
    if hasattr(module, "DeserializingConsumer"):
        wrap_function_wrapper(module, "DeserializingConsumer.__init__", wrap_DeserializingConsumer_init)
        wrap_function_wrapper(module, "DeserializingConsumer.poll", wrap_DeserializingConsumer_poll)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from testing_support.fixtures import override_application_settings, reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import \
    validate_log_event_count_outside_transaction
from testing_support.validators.validate_log_events import validate_log_events
from testing_support.validators.validate_log_events_outside_transaction import validate_log_events_outside_transaction


_common_attributes_service_linking = {"timestamp": None, "hostname": None,
                                      "entity.name": "Python Agent Test (logger_structlog)", "entity.guid": None}

_common_attributes_trace_linking = {"span.id": "abcdefgh", "trace.id": "abcdefgh12345678",
                                    **_common_attributes_service_linking}


@reset_core_stats_engine()
@override_application_settings({"application_logging.local_decorating.enabled": False})
def test_logging_inside_transaction(exercise_logging_multiple_lines):
    @validate_log_events([
        {"message": "Cat", "level": "INFO", **_common_attributes_trace_linking},
        {"message": "Dog", "level": "ERROR", **_common_attributes_trace_linking},
        {"message": "Elephant", "level": "CRITICAL", **_common_attributes_trace_linking},
    ])
    @validate_log_event_count(3)
    @background_task()
    def test():
        exercise_logging_multiple_lines()

    test()


@reset_core_stats_engine()
@override_application_settings({"application_logging.local_decorating.enabled": False})
def test_logging_filtering_inside_transaction(exercise_filtering_logging_multiple_lines):
    @validate_log_events([
        {"message": "Dog", "level": "ERROR", **_common_attributes_trace_linking},
        {"message": "Elephant", "level": "CRITICAL", **_common_attributes_trace_linking},
    ])
    @validate_log_event_count(2)
    @background_task()
    def test():
        exercise_filtering_logging_multiple_lines()

    test()


@reset_core_stats_engine()
@override_application_settings({"application_logging.local_decorating.enabled": False})
def test_logging_outside_transaction(exercise_logging_multiple_lines):
    @validate_log_events_outside_transaction([
        {"message": "Cat", "level": "INFO", **_common_attributes_service_linking},
        {"message": "Dog", "level": "ERROR", **_common_attributes_service_linking},
        {"message": "Elephant", "level": "CRITICAL", **_common_attributes_service_linking},
    ])
    @validate_log_event_count_outside_transaction(3)
    def test():
        exercise_logging_multiple_lines()

    test()


@reset_core_stats_engine()
@override_application_settings({"application_logging.local_decorating.enabled": False})
def test_logging_filtering_outside_transaction(exercise_filtering_logging_multiple_lines):
    @validate_log_events_outside_transaction([
        {"message": "Dog", "level": "ERROR", **_common_attributes_service_linking},
        {"message": "Elephant", "level": "CRITICAL", **_common_attributes_service_linking},
    ])
    @validate_log_event_count_outside_transaction(2)
    def test():
        exercise_filtering_logging_multiple_lines()

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import isawaitable


# Async Functions not allowed in Py2
async def example_middleware_async(next, root, info, **args):
    return_value = next(root, info, **args)
    if isawaitable(return_value):
        return await return_value
    return return_value


async def error_middleware_async(next, root, info, **args):
    raise RuntimeError("Runtime Error!")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from newrelic.api.application import Application, application_instance
from newrelic.api.background_task import BackgroundTask
from newrelic.api.message_trace import MessageTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.async_proxy import TransactionContext, async_proxy
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object


class MessageTransaction(BackgroundTask):
    def __init__(
        self,
        library,
        destination_type,
        destination_name,
        application,
        routing_key=None,
        exchange_type=None,
        headers=None,
        queue_name=None,
        reply_to=None,
        correlation_id=None,
        transport_type="AMQP",
        source=None,
    ):
        name, group = self.get_transaction_name(library, destination_type, destination_name)

        super(MessageTransaction, self).__init__(application, name, group=group, source=source)

        self.headers = headers

        if headers is not None and self.settings is not None:
            if self.settings.distributed_tracing.enabled:
                self.accept_distributed_trace_headers(headers, transport_type=transport_type)
            elif self.settings.cross_application_tracer.enabled:
                self._process_incoming_cat_headers(
                    headers.pop(MessageTrace.cat_id_key, None), headers.pop(MessageTrace.cat_transaction_key, None)
                )

        self.routing_key = routing_key
        self.exchange_type = exchange_type
        self.queue_name = queue_name
        self.reply_to = reply_to
        self.correlation_id = correlation_id

    @staticmethod
    def get_transaction_name(library, destination_type, destination_name):
        group = "Message/%s/%s" % (library, destination_type)
        name = "Named/%s" % destination_name
        return name, group

    def _update_agent_attributes(self):
        ms_attrs = self._agent_attributes

        if self.exchange_type is not None:
            ms_attrs["message.exchangeType"] = self.exchange_type
        if self.queue_name is not None:
            ms_attrs["message.queueName"] = self.queue_name
        if self.reply_to is not None:
            ms_attrs["message.replyTo"] = self.reply_to
        if self.correlation_id is not None:
            ms_attrs["message.correlationId"] = self.correlation_id
        if self.headers:
            for k, v in self.headers.items():
                new_key = "message.headers.%s" % k
                new_val = str(v)
                ms_attrs[new_key] = new_val
        if self.routing_key is not None:
            ms_attrs["message.routingKey"] = self.routing_key

        super(MessageTransaction, self)._update_agent_attributes()


def MessageTransactionWrapper(
    wrapped,
    library,
    destination_type,
    destination_name,
    application=None,
    routing_key=None,
    exchange_type=None,
    headers=None,
    queue_name=None,
    reply_to=None,
    correlation_id=None,
):
    def wrapper(wrapped, instance, args, kwargs):
        if callable(library):
            if instance is not None:
                _library = library(instance, *args, **kwargs)
            else:
                _library = library(*args, **kwargs)
        else:
            _library = library

        if callable(destination_type):
            if instance is not None:
                _destination_type = destination_type(instance, *args, **kwargs)
            else:
                _destination_type = destination_type(*args, **kwargs)
        else:
            _destination_type = destination_type

        if callable(destination_name):
            if instance is not None:
                _destination_name = destination_name(instance, *args, **kwargs)
            else:
                _destination_name = destination_name(*args, **kwargs)
        else:
            _destination_name = destination_name

        if callable(routing_key):
            if instance is not None:
                _routing_key = routing_key(instance, *args, **kwargs)
            else:
                _routing_key = routing_key(*args, **kwargs)
        else:
            _routing_key = routing_key

        if callable(exchange_type):
            if instance is not None:
                _exchange_type = exchange_type(instance, *args, **kwargs)
            else:
                _exchange_type = exchange_type(*args, **kwargs)
        else:
            _exchange_type = exchange_type

        if callable(headers):
            if instance is not None:
                _headers = headers(instance, *args, **kwargs)
            else:
                _headers = headers(*args, **kwargs)
        else:
            _headers = headers

        if callable(queue_name):
            if instance is not None:
                _queue_name = queue_name(instance, *args, **kwargs)
            else:
                _queue_name = queue_name(*args, **kwargs)
        else:
            _queue_name = queue_name

        if callable(reply_to):
            if instance is not None:
                _reply_to = reply_to(instance, *args, **kwargs)
            else:
                _reply_to = reply_to(*args, **kwargs)
        else:
            _reply_to = reply_to

        if callable(correlation_id):
            if instance is not None:
                _correlation_id = correlation_id(instance, *args, **kwargs)
            else:
                _correlation_id = correlation_id(*args, **kwargs)
        else:
            _correlation_id = correlation_id

        def create_transaction(transaction):
            if transaction:
                # If there is any active transaction we will return without
                # applying a new WSGI application wrapper context. In the
                # case of a transaction which is being ignored or which has
                # been stopped, we do that without doing anything further.

                if transaction.ignore_transaction or transaction.stopped:
                    return None

                if not transaction.background_task:
                    transaction.background_task = True
                    transaction.set_transaction_name(
                        *MessageTransaction.get_transaction_name(_library, _destination_type, _destination_name)
                    )

                return None

            if type(application) != Application:
                _application = application_instance(application)
            else:
                _application = application

            return MessageTransaction(
                library=_library,
                destination_type=_destination_type,
                destination_name=_destination_name,
                application=_application,
                routing_key=_routing_key,
                exchange_type=_exchange_type,
                headers=_headers,
                queue_name=_queue_name,
                reply_to=_reply_to,
                correlation_id=_correlation_id,
                source=wrapped,
            )

        proxy = async_proxy(wrapped)
        if proxy:
            context_manager = TransactionContext(create_transaction)
            return proxy(wrapped(*args, **kwargs), context_manager)

        manager = create_transaction(current_transaction(active_only=False))

        # This means that transaction already exists and we want to return
        if not manager:
            return wrapped(*args, **kwargs)

        with manager:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def message_transaction(
    library,
    destination_type,
    destination_name,
    application=None,
    routing_key=None,
    exchange_type=None,
    headers=None,
    queue_name=None,
    reply_to=None,
    correlation_id=None,
):
    return functools.partial(
        MessageTransactionWrapper,
        library=library,
        destination_type=destination_type,
        destination_name=destination_name,
        application=application,
        routing_key=routing_key,
        exchange_type=exchange_type,
        headers=headers,
        queue_name=queue_name,
        reply_to=reply_to,
        correlation_id=correlation_id,
    )


def wrap_message_transaction(
    module,
    object_path,
    library,
    destination_type,
    destination_name,
    application=None,
    routing_key=None,
    exchange_type=None,
    headers=None,
    queue_name=None,
    reply_to=None,
    correlation_id=None,
):
    wrap_object(
        module,
        object_path,
        MessageTransactionWrapper,
        (
            library,
            destination_type,
            destination_name,
            application,
            routing_key,
            exchange_type,
            headers,
            queue_name,
            reply_to,
            correlation_id,
        ),
    )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

import cherrypy

class Resource(object):

    exposed = True

    def GET(self):
        return 'GET RESPONSE'

dispatcher = cherrypy.dispatch.MethodDispatcher()

conf = { '/': { 'request.dispatch': dispatcher } }

application = cherrypy.Application(Resource(), '/', conf)
test_application = webtest.TestApp(application)

@validate_code_level_metrics("test_resource.Resource", "GET")
@validate_transaction_errors(errors=[])
def test_resource_get():
    response = test_application.get('')
    response.mustcontain('GET RESPONSE')

@validate_transaction_errors(errors=[])
def test_resource_not_found():
    response = test_application.post('', status=405)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import aiohttp
import pytest
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from yarl import URL

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import function_trace

version_info = tuple(int(_) for _ in aiohttp.__version__.split(".")[:2])
skipif_aiohttp3 = pytest.mark.skipif(
    version_info >= (3, 0), reason="This version of aiohttp does not support yield from syntax"
)


async def fetch(method, url):
    with aiohttp.ClientSession() as session:
        _method = getattr(session, method)
        response = await asyncio.wait_for(_method(url), timeout=None)
        response.raise_for_status()
        await response.text()


@background_task(name="fetch_multiple")
async def fetch_multiple(method, url):
    coros = [fetch(method, url) for _ in range(2)]
    return asyncio.gather(*coros, return_exceptions=True)


if version_info < (2, 0):
    _expected_error_class = aiohttp.errors.HttpProcessingError
else:
    _expected_error_class = aiohttp.client_exceptions.ClientResponseError


def task(loop, method, exc_expected, url):
    future = asyncio.ensure_future(fetch_multiple(method, url))
    text_list = loop.run_until_complete(future)
    if exc_expected:
        assert isinstance(text_list[0], _expected_error_class)
        assert isinstance(text_list[1], _expected_error_class)
    else:
        assert text_list[0] == text_list[1]


test_matrix = (
    ("get", False),
    ("post", True),
    ("options", True),
    ("head", True),
    ("put", True),
    ("patch", True),
    ("delete", True),
)


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_client_yield_from(event_loop, local_server_info, method, exc_expected):
    @validate_transaction_metrics(
        "fetch_multiple",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
    )
    def task_test():
        task(event_loop, method, exc_expected, local_server_info.url)

    task_test()


@skipif_aiohttp3
def test_client_yarl_yield_from(event_loop, local_server_info):
    method = "get"

    @validate_transaction_metrics(
        "fetch_multiple",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
    )
    def task_test():
        task(event_loop, method, False, URL(local_server_info.url))

    task_test()


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_client_no_txn_yield_from(event_loop, local_server_info, method, exc_expected):
    def task_test():
        task(event_loop, method, exc_expected, local_server_info.url)

    task_test()


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_client_throw_yield_from(event_loop, local_server_info, method, exc_expected):
    class ThrowerException(ValueError):
        pass

    @background_task(name="test_client_throw_yield_from")
    async def self_driving_thrower():
        with aiohttp.ClientSession() as session:
            coro = session._request(method.upper(), local_server_info.url)

            # activate the coroutine
            coro.send(None)

            # inject error
            coro.throw(ThrowerException())

    @validate_transaction_metrics(
        "test_client_throw_yield_from",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + method.upper(), 1),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + method.upper(), 1),
        ],
    )
    def task_test():

        with pytest.raises(ThrowerException):
            event_loop.run_until_complete(self_driving_thrower())

    task_test()


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_client_close_yield_from(event_loop, local_server_info, method, exc_expected):
    @background_task(name="test_client_close_yield_from")
    async def self_driving_closer():
        with aiohttp.ClientSession() as session:
            coro = session._request(method.upper(), local_server_info.url)

            # activate the coroutine
            coro.send(None)

            # force close
            coro.close()

    @validate_transaction_metrics(
        "test_client_close_yield_from",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + method.upper(), 1),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + method.upper(), 1),
        ],
    )
    def task_test():
        event_loop.run_until_complete(self_driving_closer())

    task_test()


test_ws_matrix = (
    # the 127.0.0.1 server does not accept websocket requests, hence an
    # exception is expected but a metric will still be created
    ("ws_connect", True),
)


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_ws_matrix)
def test_ws_connect_yield_from(event_loop, local_server_info, method, exc_expected):
    @validate_transaction_metrics(
        "fetch_multiple",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + "GET", 2),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + "GET", 2),
        ],
    )
    def task_test():
        task(event_loop, method, exc_expected, local_server_info.url)

    task_test()


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_create_task_yield_from(event_loop, local_server_info, method, exc_expected):

    # `loop.create_task` returns a Task object which uses the coroutine's
    # `send` method, not `__next__`

    async def fetch_task(loop):
        with aiohttp.ClientSession() as session:
            coro = getattr(session, method)
            resp = await loop.create_task(coro(local_server_info.url))
            resp.raise_for_status()
            await resp.text()

    @background_task(name="test_create_task_yield_from")
    async def fetch_multiple(loop):
        coros = [fetch_task(loop) for _ in range(2)]
        return asyncio.gather(*coros, return_exceptions=True)

    @validate_transaction_metrics(
        "test_create_task_yield_from",
        background_task=True,
        scoped_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
        rollup_metrics=[
            (local_server_info.base_metric + method.upper(), 2),
        ],
    )
    def task_test():
        result = event_loop.run_until_complete(fetch_multiple(event_loop))
        if exc_expected:
            assert isinstance(result[0], _expected_error_class)
            assert isinstance(result[1], _expected_error_class)
        else:
            assert result[0] == result[1]

    task_test()


@skipif_aiohttp3
@pytest.mark.parametrize("method,exc_expected", test_matrix)
def test_terminal_node_yield_from(event_loop, local_server_info, method, exc_expected):
    """
    This test injects a terminal node into a simple background task workflow.
    It was added to validate a bug where our coro.send() wrapper would fail
    when transaction's current node was terminal.
    """

    def task_test():
        @function_trace(terminal=True)
        def execute_task():
            task(event_loop, method, exc_expected, local_server_info.url)

        execute_task()

    task_test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from django.conf.urls.defaults import url
except ImportError:
    try:
        from django.conf.urls import url
    except ImportError:
        from django.urls import re_path as url

import views

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^exception$', views.exception, name='exception'),
    url(r'^middleware_410$', views.middleware_410, name='middleware_410'),
    url(r'^permission_denied$', views.permission_denied,
            name='permission_denied'),
    url(r'^cbv$', views.MyView.as_view()),
    url(r'^deferred_cbv$', views.deferred_cbv),
    url(r'^html_insertion$', views.html_insertion, name='html_insertion'),
    url(r'^html_insertion_content_length$',
        views.html_insertion_content_length,
        name='html_insertion_content_length'),
    url(r'^html_insertion_manual$', views.html_insertion_manual,
        name='html_insertion_manual'),
    url(r'^html_insertion_unnamed_attachment_header$',
        views.html_insertion_unnamed_attachment_header,
        name='html_insertion_unnamed_attachment_header'),
    url(r'^html_insertion_named_attachment_header$',
        views.html_insertion_named_attachment_header,
        name='html_insertion_named_attachment_header'),
    url(r'^inclusion_tag$', views.inclusion_tag, name='inclusion_tag'),
    url(r'^template_tags$', views.template_tags, name='template_tags'),
    url(r'^render_exception_function', views.render_exception_function,
        name='render_exception_function'),
    url(r'^render_exception_class', views.RenderExceptionClass.as_view(),
        name='render_exception_class'),
    url(r'^gzip_html_insertion', views.gzip_html_insertion,
        name='gzip_html_insertion'),
]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from newrelic.packages import six

from newrelic.common.object_wrapper import (transient_function_wrapper,
        function_wrapper)
from testing_support.fixtures import catch_background_exceptions

def validate_log_events(events):
    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        record_called = []
        recorded_logs = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        @catch_background_exceptions
        def _validate_log_events(wrapped, instance, args, kwargs):
            record_called.append(True)
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            else:
                recorded_logs[:] = []
                recorded_logs.extend(list(instance._log_events))

            return result


        _new_wrapper = _validate_log_events(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert record_called
        logs = copy.copy(recorded_logs)
        
        record_called[:] = []
        recorded_logs[:] = []

        for expected in events:
            matching_log_events = 0
            mismatches = []
            for captured in logs:
                if _check_log_attributes(expected, captured, mismatches):
                    matching_log_events += 1
            assert matching_log_events == 1, _log_details(matching_log_events, logs, mismatches)

        return val


    def _check_log_attributes(expected, captured, mismatches):
        for key, value in six.iteritems(expected):
            if hasattr(captured, key):
                captured_value = getattr(captured, key, None)
            elif key in captured.attributes:
                captured_value = captured.attributes[key]
            else:
                mismatches.append("key: %s, value:<%s><%s>" % (key, value, getattr(captured, key, None)))
                return False

            if value is not None:
                if value != captured_value:
                    mismatches.append("key: %s, value:<%s><%s>" % (key, value, captured_value))
                    return False

        return True

    def _log_details(matching_log_events, captured, mismatches):
        details = [
            "matching_log_events=%d" % matching_log_events,
            "mismatches=%s" % mismatches,
            "captured_events=%s" % captured,
        ]

        return "\n".join(details)

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_dimensional_metric_payload import (
    validate_dimensional_metric_payload,
)
from testing_support.validators.validate_metric_payload import validate_metric_payload

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.transaction import record_custom_metric, record_dimensional_metric
from newrelic.core.rules_engine import NormalizationRule, RulesEngine

RULES = [{"match_expression": "(replace)", "replacement": "expected", "ignore": False, "eval_order": 0}]
EXPECTED_TAGS = frozenset({"tag": 1}.items())


def _prepare_rules(test_rules):
    # ensure all keys are present, if not present set to an empty string
    for rule in test_rules:
        for key in NormalizationRule._fields:
            rule[key] = rule.get(key, "")
    return test_rules


@pytest.fixture(scope="session")
def core_app(collector_agent_registration):
    app = collector_agent_registration
    return app._agent.application(app.name)


@pytest.fixture(scope="function")
def rules_engine_fixture(core_app):
    rules_engine = core_app._rules_engine
    previous_rules = rules_engine["metric"]

    rules_engine["metric"] = RulesEngine(_prepare_rules(RULES))
    yield
    rules_engine["metric"] = previous_rules  # Restore after test run


@validate_dimensional_metric_payload(summary_metrics=[("Metric/expected", EXPECTED_TAGS, 1)])
@validate_metric_payload([("Metric/expected", 1)])
@reset_core_stats_engine()
def test_metric_normalization_inside_transaction(core_app, rules_engine_fixture):
    @background_task(name="test_record_dimensional_metric_inside_transaction")
    def _test():
        record_dimensional_metric("Metric/replace", 1, tags={"tag": 1})
        record_custom_metric("Metric/replace", 1)

    _test()
    core_app.harvest()


@validate_dimensional_metric_payload(summary_metrics=[("Metric/expected", EXPECTED_TAGS, 1)])
@validate_metric_payload([("Metric/expected", 1)])
@reset_core_stats_engine()
def test_metric_normalization_outside_transaction(core_app, rules_engine_fixture):
    def _test():
        app = application_instance()
        record_dimensional_metric("Metric/replace", 1, tags={"tag": 1}, application=app)
        record_custom_metric("Metric/replace", 1, application=app)

    _test()
    core_app.harvest()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc._channel
import pytest

from newrelic.hooks.framework_grpc import _get_uri_method


_test_get_url_unary_unary = [
        ('localhost:1234', '/sample/method',
            'grpc://localhost:1234/sample/method'),
        ('localhost:1234', 'method/without/leading/slash',
            'grpc://localhost:1234/method/without/leading/slash'),
        ('localhost', '/no/port',
            'grpc://localhost/no/port'),
]

_test_channel_types = [
        ('unary_unary', grpc._channel._UnaryUnaryMultiCallable),
        ('unary_stream', grpc._channel._UnaryStreamMultiCallable),
        ('stream_unary', grpc._channel._StreamUnaryMultiCallable),
        ('stream_stream', grpc._channel._StreamStreamMultiCallable),
]


@pytest.mark.parametrize('url,method,expected', _test_get_url_unary_unary)
@pytest.mark.parametrize('channel_type,channel_class', _test_channel_types)
def test_get_url_method(url, method, expected, channel_type,
        channel_class):
    channel = grpc.insecure_channel(url)
    unary_unary = getattr(channel, channel_type)(method)
    assert type(unary_unary) == channel_class

    actual_url, actual_method = _get_uri_method(unary_unary)
    assert actual_url == expected
    assert actual_method == method.lstrip('/')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import pytest
from testing_support.fixtures import override_generic_settings

from newrelic.common.streaming_utils import StreamBuffer
from newrelic.core.agent_streaming import StreamingRpc
from newrelic.core.config import global_settings
from newrelic.core.infinite_tracing_pb2 import AttributeValue, Span

CONDITION_CLS = type(threading.Condition())
DEFAULT_METADATA = (("agent_run_token", ""), ("license_key", ""))


def record_metric(*args, **kwargs):
    pass


# This enumeration is taken from gRPC's implementation for compression:
# https://grpc.github.io/grpc/python/grpc.html#compression
@pytest.mark.parametrize(
    "compression_setting, gRPC_compression_val",
    (
        (None, 0),
        (True, 2),
        (False, 0),
    ),
)
def test_correct_settings(mock_grpc_server, compression_setting, gRPC_compression_val):
    settings = global_settings()

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.compression": compression_setting,
        },
    )
    def _test():
        endpoint = "localhost:%s" % mock_grpc_server
        stream_buffer = StreamBuffer(1)

        rpc = StreamingRpc(
            endpoint,
            stream_buffer,
            DEFAULT_METADATA,
            record_metric,
            ssl=False,
            compression=settings.infinite_tracing.compression,
        )

        rpc.connect()
        assert rpc.compression_setting.value == gRPC_compression_val
        rpc.close()

    _test()


def test_close_before_connect(mock_grpc_server, batching):
    endpoint = "localhost:%s" % mock_grpc_server
    stream_buffer = StreamBuffer(0, batching=batching)

    rpc = StreamingRpc(endpoint, stream_buffer, DEFAULT_METADATA, record_metric, ssl=False)

    # Calling close will close the grpc channel
    rpc.close()
    rpc.connect()
    # The response processing thread should immediately exit if the channel is
    # closed
    rpc.response_processing_thread.join(timeout=5)
    assert not rpc.response_processing_thread.is_alive()


def test_close_while_connected(mock_grpc_server, buffer_empty_event, batching):
    endpoint = "localhost:%s" % mock_grpc_server
    stream_buffer = StreamBuffer(1, batching=batching)

    rpc = StreamingRpc(endpoint, stream_buffer, DEFAULT_METADATA, record_metric, ssl=False)

    rpc.connect()
    # Check the processing thread is alive and spans are being sent
    assert rpc.response_processing_thread.is_alive()

    span = Span(intrinsics={}, agent_attributes={}, user_attributes={})

    buffer_empty_event.clear()
    stream_buffer.put(span)

    assert buffer_empty_event.wait(5)

    rpc.close()
    assert not rpc.response_processing_thread.is_alive()


def test_close_while_awaiting_reconnect(mock_grpc_server, monkeypatch, batching):
    event = threading.Event()

    class WaitOnWait(CONDITION_CLS):
        def wait(self, *args, **kwargs):
            event.set()
            # Call super wait with no arguments to block until a notify
            return super(WaitOnWait, self).wait()

    @staticmethod
    def condition(*args, **kwargs):
        return WaitOnWait(*args, **kwargs)

    monkeypatch.setattr(StreamingRpc, "condition", condition)

    span = Span(
        intrinsics={"status_code": AttributeValue(string_value="INTERNAL")},
        agent_attributes={},
        user_attributes={},
    )

    endpoint = "localhost:%s" % mock_grpc_server
    stream_buffer = StreamBuffer(1, batching=batching)

    rpc = StreamingRpc(endpoint, stream_buffer, DEFAULT_METADATA, record_metric, ssl=False)

    rpc.connect()
    # Send a span to trigger reconnect
    stream_buffer.put(span)
    # Wait until for StreamingRpc to pause before attempting reconnect
    assert event.wait(timeout=5)
    # Close the rpc
    rpc.close()
    # Make sure the processing_thread is closed
    assert not rpc.response_processing_thread.is_alive()


@pytest.mark.parametrize("compression", (True, False))
def test_rpc_serialization_and_deserialization(
    mock_grpc_server,
    batching,
    compression,
    buffer_empty_event,
    spans_received,
    span_batches_received,
    spans_processed_event,
):
    """StreamingRPC sends deserializable span to correct endpoint."""

    endpoint = "localhost:%s" % mock_grpc_server
    stream_buffer = StreamBuffer(1, batching=batching)

    span = Span(
        intrinsics={},
        agent_attributes={},
        user_attributes={},
    )

    rpc = StreamingRpc(endpoint, stream_buffer, DEFAULT_METADATA, record_metric, compression=compression, ssl=False)

    rpc.connect()

    buffer_empty_event.clear()
    stream_buffer.put(span)

    assert buffer_empty_event.wait(5)
    assert spans_processed_event.wait(5)

    if batching:
        assert not spans_received, "Spans incorrectly received."
        assert span_batches_received, "No span batches received."
    else:
        assert not span_batches_received, "Span batches incorrectly received."
        assert spans_received, "No spans received."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import sys
import tempfile

import pytest

# NOTE: the test_utilization_settings_from_env_vars test mocks several of the
# methods in newrelic.core.data_collector and does not put them back!
from testing_support.mock_http_client import create_client_cls

import newrelic.core.config
from newrelic.common.object_wrapper import function_wrapper
from newrelic.common.system_info import BootIdUtilization
from newrelic.common.utilization import CommonUtilization
from newrelic.core.agent_protocol import AgentProtocol

try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload

INITIAL_ENV = os.environ

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures", "utilization", "utilization_json.json"))


def _load_tests():
    with open(FIXTURE, "r") as fh:
        js = fh.read()
    return json.loads(js)


def _mock_logical_processor_count(cnt):
    def logical_processor_count():
        return cnt

    return logical_processor_count


def _mock_total_physical_memory(mem):
    def total_physical_memory():
        return mem

    return total_physical_memory


def _mock_gethostname(name):
    def gethostname(*args, **kwargs):
        return name

    return gethostname


def _mock_getips(ip_addresses):
    def getips(*args, **kwargs):
        return ip_addresses

    return getips


class UpdatedSettings(object):
    def __init__(self):
        self.initial_settings = newrelic.core.config._settings

    def __enter__(self):
        """Update the settings dict to reflect the environment variables found in
        the test.

        """
        # clean settings cache and reload env vars
        # Note that reload can at times work in unexpected ways. All that is
        # required here is that the globals (such as
        # newrelic.core.config._settings) be reset.
        #
        # From python docs (2.x and 3.x)
        # "When a module is reloaded, its dictionary (containing the module's
        # global variables) is retained. Redefinitions of names will override
        # the old definitions, so this is generally not a problem."
        reload(newrelic.core.config)
        reload(newrelic.config)

        return newrelic.core.config.global_settings()

    def __exit__(self, *args, **kwargs):
        newrelic.core.config._settings = self.initial_settings


def _get_response_body_for_test(test):
    if test.get("input_aws_id"):
        return json.dumps(
            {
                "instanceId": test.get("input_aws_id"),
                "instanceType": test.get("input_aws_type"),
                "availabilityZone": test.get("input_aws_zone"),
            }
        ).encode("utf8")
    if test.get("input_azure_id"):
        return json.dumps(
            {
                "location": test.get("input_azure_location"),
                "name": test.get("input_azure_name"),
                "vmId": test.get("input_azure_id"),
                "vmSize": test.get("input_azure_size"),
            }
        ).encode("utf8")
    if test.get("input_gcp_id"):
        return json.dumps(
            {
                "id": test.get("input_gcp_id"),
                "machineType": test.get("input_gcp_type"),
                "name": test.get("input_gcp_name"),
                "zone": test.get("input_gcp_zone"),
            }
        ).encode("utf8")


def patch_boot_id_file(test):
    @function_wrapper
    def _patch_boot_id_file(wrapped, instance, args, kwargs):
        boot_id_file = None
        initial_sys_platform = sys.platform

        if test.get("input_boot_id"):
            boot_id_file = tempfile.NamedTemporaryFile()
            boot_id_file.write(test.get("input_boot_id"))
            boot_id_file.seek(0)
            BootIdUtilization.METADATA_URL = boot_id_file.name
            sys.platform = "linux-mock-testing"  # ensure boot_id is gathered
        else:
            # do not gather boot_id at all, this will ensure there is nothing
            # extra in the gathered utilizations data
            sys.platform = "not-linux"

        try:
            return wrapped(*args, **kwargs)
        finally:
            del boot_id_file  # close and thus delete the tempfile
            sys.platform = initial_sys_platform

    return _patch_boot_id_file


def patch_system_info(test, monkeypatch):
    @function_wrapper
    def _patch_system_info(wrapped, instance, args, kwargs):
        sys_info = newrelic.common.system_info

        monkeypatch.setattr(
            sys_info, "logical_processor_count", _mock_logical_processor_count(test.get("input_logical_processors"))
        )
        monkeypatch.setattr(
            sys_info, "total_physical_memory", _mock_total_physical_memory(test.get("input_total_ram_mib"))
        )
        monkeypatch.setattr(sys_info, "gethostname", _mock_gethostname(test.get("input_hostname")))
        monkeypatch.setattr(sys_info, "getips", _mock_getips(test.get("input_ip_address")))

        return wrapped(*args, **kwargs)

    return _patch_system_info


@pytest.mark.parametrize("test", _load_tests())
def test_utilization_settings(test, monkeypatch):

    env = test.get("input_environment_variables", {})

    if test.get("input_pcf_guid"):
        env.update(
            {
                "CF_INSTANCE_GUID": test.get("input_pcf_guid"),
                "CF_INSTANCE_IP": test.get("input_pcf_ip"),
                "MEMORY_LIMIT": test.get("input_pcf_mem_limit"),
            }
        )

    for key, val in env.items():
        monkeypatch.setenv(key, str(val))

    @patch_boot_id_file(test)
    @patch_system_info(test, monkeypatch)
    def _test_utilization_data():

        data = _get_response_body_for_test(test)
        client_cls = create_client_cls(200, data)
        monkeypatch.setattr(CommonUtilization, "CLIENT_CLS", client_cls)

        with UpdatedSettings() as settings:
            # Ignoring docker will ensure that there is nothing extra in the
            # gathered utilization data
            monkeypatch.setattr(settings.utilization, "detect_docker", False)

            (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
            util_output = local_config["utilization"]
            expected_output = test["expected_output_json"]

            # The agent does not record full_hostname and it's not required
            expected_output.pop("full_hostname")

            assert expected_output == util_output

    _test_utilization_data()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cherrypy
import pytest
import webtest
from testing_support.fixtures import (
    override_application_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)

from newrelic.packages import six

CHERRYPY_VERSION = tuple(int(v) for v in cherrypy.__version__.split("."))


class Application(object):
    @cherrypy.expose
    def index(self):
        return "INDEX RESPONSE"

    @cherrypy.expose
    def error(self):
        raise RuntimeError("error")

    @cherrypy.expose
    def not_found(self):
        raise cherrypy.NotFound()

    @cherrypy.expose
    def not_found_as_http_error(self):
        raise cherrypy.HTTPError(404)

    @cherrypy.expose
    def not_found_as_str_http_error(self):
        raise cherrypy.HTTPError("404 Not Found")

    @cherrypy.expose
    def bad_http_error(self):
        # this will raise HTTPError with status code 500 because 10 is not a
        # valid status code
        raise cherrypy.HTTPError("10 Invalid status code")

    @cherrypy.expose
    def internal_redirect(self):
        raise cherrypy.InternalRedirect("/")

    @cherrypy.expose
    def external_redirect(self):
        raise cherrypy.HTTPRedirect("/")

    @cherrypy.expose
    def upload_files(self, files):
        return "UPLOAD FILES RESPONSE"

    @cherrypy.expose
    def encode_multipart(self, field, files):
        return "ENCODE MULTIPART RESPONSE"

    @cherrypy.expose
    def html_insertion(self):
        return (
            "<!DOCTYPE html><html><head>Some header</head>"
            "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
            "</body></html>"
        )


application = cherrypy.Application(Application())
test_application = webtest.TestApp(application)


@validate_code_level_metrics("test_application.Application", "index")
@validate_transaction_errors(errors=[])
def test_application_index():
    response = test_application.get("")
    response.mustcontain("INDEX RESPONSE")


@validate_transaction_errors(errors=[])
def test_application_index_agent_disabled():
    environ = {"newrelic.enabled": False}
    response = test_application.get("", extra_environ=environ)
    response.mustcontain("INDEX RESPONSE")


@validate_transaction_errors(errors=[])
def test_application_missing():
    test_application.get("/missing", status=404)


if six.PY3:
    _test_application_unexpected_exception_errors = ["builtins:RuntimeError"]
else:
    _test_application_unexpected_exception_errors = ["exceptions:RuntimeError"]


@validate_transaction_errors(errors=_test_application_unexpected_exception_errors)
def test_application_unexpected_exception():
    test_application.get("/error", status=500)


@validate_transaction_errors(errors=[])
def test_application_not_found():
    test_application.get("/not_found", status=404)


@validate_transaction_errors(errors=[])
def test_application_not_found_as_http_error():
    test_application.get("/not_found_as_http_error", status=404)


@validate_transaction_errors(errors=[])
def test_application_internal_redirect():
    response = test_application.get("/internal_redirect")
    response.mustcontain("INDEX RESPONSE")


@validate_transaction_errors(errors=[])
def test_application_external_redirect():
    test_application.get("/external_redirect", status=302)


@validate_transaction_errors(errors=[])
def test_application_upload_files():
    test_application.post("/upload_files", upload_files=[("files", __file__)])


@validate_transaction_errors(errors=[])
def test_application_encode_multipart():
    content_type, body = test_application.encode_multipart(params=[("field", "value")], files=[("files", __file__)])
    test_application.request("/encode_multipart", method="POST", content_type=content_type, body=body)


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion():
    response = test_application.get("/html_insertion")

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


_error_endpoints = ["/not_found_as_http_error"]
if CHERRYPY_VERSION >= (3, 2):
    _error_endpoints.extend(["/not_found_as_str_http_error", "/bad_http_error"])


@pytest.mark.parametrize("endpoint", _error_endpoints)
@pytest.mark.parametrize(
    "ignore_overrides,expected_errors",
    [
        ([], ["cherrypy._cperror:HTTPError"]),
        ([404, 500], []),
    ],
)
def test_ignore_status_code(endpoint, ignore_overrides, expected_errors):
    @validate_transaction_errors(errors=expected_errors)
    @override_ignore_status_codes(ignore_overrides)
    def _test():
        test_application.get(endpoint, status=[404, 500])

    _test()


@validate_transaction_errors(errors=[])
def test_ignore_status_unexpected_param():
    response = test_application.get("/?arg=1", status=404)
    response.mustcontain(no=["INDEX RESPONSE"])



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sklearn import __init__  # noqa: Needed for get_package_version
from sklearn.ensemble import AdaBoostClassifier
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version
from newrelic.packages import six

SKLEARN_VERSION = tuple(map(int, get_package_version("sklearn").split(".")))


# Python 2 will not allow instantiation of abstract class
# (abstract method is __init__ here)
@pytest.mark.skipif(SKLEARN_VERSION >= (1, 0, 0) or six.PY2, reason="Requires sklearn < 1.0 and Python3")
@pytest.mark.parametrize(
    "multioutput_model_name",
    [
        "MultiOutputEstimator",
    ],
)
def test_below_v1_0_model_methods_wrapped_in_function_trace(multioutput_model_name, run_multioutput_model):
    expected_scoped_metrics = {
        "MultiOutputEstimator": [
            ("Function/MLModel/Sklearn/Named/MultiOutputEstimator.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiOutputEstimator.predict", 2),
        ],
    }
    expected_transaction_name = (
        "test_multioutput_models:test_below_v1_0_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_multioutput_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[multioutput_model_name],
        rollup_metrics=expected_scoped_metrics[multioutput_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_multioutput_model(multioutput_model_name)

    _test()


@pytest.mark.parametrize(
    "multioutput_model_name",
    [
        "MultiOutputClassifier",
        "ClassifierChain",
        "RegressorChain",
    ],
)
def test_above_v1_0_model_methods_wrapped_in_function_trace(multioutput_model_name, run_multioutput_model):
    expected_scoped_metrics = {
        "MultiOutputClassifier": [
            ("Function/MLModel/Sklearn/Named/MultiOutputClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiOutputClassifier.predict_proba", 1),
            ("Function/MLModel/Sklearn/Named/MultiOutputClassifier.score", 1),
        ],
        "ClassifierChain": [
            ("Function/MLModel/Sklearn/Named/ClassifierChain.fit", 1),
            ("Function/MLModel/Sklearn/Named/ClassifierChain.predict_proba", 1),
        ],
        "RegressorChain": [
            ("Function/MLModel/Sklearn/Named/RegressorChain.fit", 1),
        ],
    }
    expected_transaction_name = (
        "test_multioutput_models:test_above_v1_0_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_multioutput_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[multioutput_model_name],
        rollup_metrics=expected_scoped_metrics[multioutput_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_multioutput_model(multioutput_model_name)

    _test()


@pytest.fixture
def run_multioutput_model():
    def _run(multioutput_model_name):
        import sklearn.multioutput
        from sklearn.datasets import make_multilabel_classification

        X, y = make_multilabel_classification(n_classes=3, random_state=0)

        kwargs = {"estimator": AdaBoostClassifier()}
        if multioutput_model_name in ["RegressorChain", "ClassifierChain"]:
            kwargs = {"base_estimator": AdaBoostClassifier()}
        clf = getattr(sklearn.multioutput, multioutput_model_name)(**kwargs)

        model = clf.fit(X, y)
        if hasattr(model, "predict"):
            model.predict(X)
        if hasattr(model, "score"):
            model.score(X, y)
        if hasattr(model, "predict_proba"):
            model.predict_proba(X)

        return model

    return _run
